#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime
from unittest import mock

import pytest
import time_machine
from moto import mock_aws

from airflow.exceptions import AirflowException
from airflow.models import DAG, DagRun, TaskInstance
from airflow.models.variable import Variable
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.sensors.s3 import S3KeySensor, S3KeysUnchangedSensor
from airflow.utils import timezone
from airflow.utils.state import DagRunState
from airflow.utils.types import DagRunType

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

DEFAULT_DATE = datetime(2015, 1, 1)


class TestS3KeySensor:
    def test_bucket_name_none_and_bucket_key_as_relative_path(self):
        """
        Test if exception is raised when bucket_name is None
        and bucket_key is provided as relative path rather than s3:// url.
        :return:
        """
        op = S3KeySensor(task_id="s3_key_sensor", bucket_key="file_in_bucket")
        with pytest.raises(AirflowException):
            op.poke(None)

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    def test_bucket_name_none_and_bucket_key_is_list_and_contain_relative_path(self, mock_head_object):
        """
        Test if exception is raised when bucket_name is None
        and bucket_key is provided with one of the two keys as relative path rather than s3:// url.
        :return:
        """
        mock_head_object.return_value = {"ContentLength": 0}
        op = S3KeySensor(task_id="s3_key_sensor", bucket_key=["s3://test_bucket/file", "file_in_bucket"])
        with pytest.raises(AirflowException):
            op.poke(None)

    def test_bucket_name_provided_and_bucket_key_is_s3_url(self):
        """
        Test if exception is raised when bucket_name is provided
        while bucket_key is provided as a full s3:// url.
        :return:
        """
        op = S3KeySensor(
            task_id="s3_key_sensor", bucket_key="s3://test_bucket/file", bucket_name="test_bucket"
        )
        with pytest.raises(TypeError):
            op.poke(None)

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    def test_bucket_name_provided_and_bucket_key_is_list_and_contains_s3_url(self, mock_head_object):
        """
        Test if exception is raised when bucket_name is provided
        while bucket_key contains a full s3:// url.
        :return:
        """
        mock_head_object.return_value = {"ContentLength": 0}
        op = S3KeySensor(
            task_id="s3_key_sensor",
            bucket_key=["test_bucket", "s3://test_bucket/file"],
            bucket_name="test_bucket",
        )
        with pytest.raises(TypeError):
            op.poke(None)

    @pytest.mark.parametrize(
        "key, bucket, parsed_key, parsed_bucket",
        [
            pytest.param("s3://bucket/key", None, "key", "bucket", id="key as s3url"),
            pytest.param("key", "bucket", "key", "bucket", id="separate bucket and key"),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    def test_parse_bucket_key(self, mock_head_object, key, bucket, parsed_key, parsed_bucket):
        print(key, bucket, parsed_key, parsed_bucket)
        mock_head_object.return_value = None

        op = S3KeySensor(
            task_id="s3_key_sensor",
            bucket_key=key,
            bucket_name=bucket,
        )

        op.poke(None)

        mock_head_object.assert_called_once_with(parsed_key, parsed_bucket)

    @pytest.mark.db_test
    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    def test_parse_bucket_key_from_jinja(self, mock_head_object, session, clean_dags_and_dagruns):
        mock_head_object.return_value = None

        Variable.set("test_bucket_key", "s3://bucket/key", session=session)

        logical_date = timezone.datetime(2020, 1, 1)

        dag = DAG("test_s3_key", schedule=None, start_date=logical_date)
        op = S3KeySensor(
            task_id="s3_key_sensor",
            bucket_key="{{ var.value.test_bucket_key }}",
            bucket_name=None,
            dag=dag,
        )

        if AIRFLOW_V_3_0_PLUS:
            dag_run = DagRun(
                dag_id=dag.dag_id,
                logical_date=logical_date,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        else:
            dag_run = DagRun(
                dag_id=dag.dag_id,
                execution_date=logical_date,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        ti = TaskInstance(task=op)
        ti.dag_run = dag_run
        session.add(ti)
        session.commit()
        context = ti.get_template_context(session)
        ti.render_templates(context)
        op.poke(None)

        mock_head_object.assert_called_once_with("key", "bucket")

    @pytest.mark.db_test
    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    def test_parse_list_of_bucket_keys_from_jinja(self, mock_head_object, session, clean_dags_and_dagruns):
        mock_head_object.return_value = None
        mock_head_object.side_effect = [{"ContentLength": 0}, {"ContentLength": 0}]

        Variable.set("test_bucket_key", ["s3://bucket/file1", "s3://bucket/file2"], session=session)

        logical_date = timezone.datetime(2020, 1, 1)

        dag = DAG("test_s3_key", schedule=None, start_date=logical_date, render_template_as_native_obj=True)
        op = S3KeySensor(
            task_id="s3_key_sensor",
            bucket_key="{{ var.value.test_bucket_key }}",
            bucket_name=None,
            dag=dag,
        )
        if hasattr(DagRun, "execution_date"):  # Airflow 2.x.
            dag_run = DagRun(
                dag_id=dag.dag_id,
                execution_date=logical_date,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        else:
            dag_run = DagRun(
                dag_id=dag.dag_id,
                logical_date=logical_date,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        ti = TaskInstance(task=op)
        ti.dag_run = dag_run
        session.add(ti)
        session.commit()
        context = ti.get_template_context(session)
        ti.render_templates(context)
        op.poke(None)

        mock_head_object.assert_any_call("file1", "bucket")
        mock_head_object.assert_any_call("file2", "bucket")

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    def test_poke(self, mock_head_object):
        op = S3KeySensor(task_id="s3_key_sensor", bucket_key="s3://test_bucket/file")

        mock_head_object.return_value = None
        assert op.poke(None) is False
        mock_head_object.assert_called_once_with("file", "test_bucket")

        mock_head_object.return_value = {"ContentLength": 0}
        assert op.poke(None) is True

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    def test_poke_multiple_files(self, mock_head_object):
        op = S3KeySensor(
            task_id="s3_key_sensor", bucket_key=["s3://test_bucket/file1", "s3://test_bucket/file2"]
        )

        mock_head_object.side_effect = [{"ContentLength": 0}, None]
        assert op.poke(None) is False

        mock_head_object.side_effect = [{"ContentLength": 0}, {"ContentLength": 0}]
        assert op.poke(None) is True

        mock_head_object.assert_any_call("file1", "test_bucket")
        mock_head_object.assert_any_call("file2", "test_bucket")

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.get_file_metadata")
    def test_poke_wildcard(self, mock_get_file_metadata):
        op = S3KeySensor(task_id="s3_key_sensor", bucket_key="s3://test_bucket/file*", wildcard_match=True)

        mock_get_file_metadata.return_value = []
        assert op.poke(None) is False
        mock_get_file_metadata.assert_called_once_with("file", "test_bucket")

        mock_get_file_metadata.return_value = [{"Key": "dummyFile", "Size": 0}]
        assert op.poke(None) is False

        mock_get_file_metadata.return_value = [{"Key": "file1", "Size": 0}]
        assert op.poke(None) is True

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.get_file_metadata")
    def test_poke_wildcard_multiple_files(self, mock_get_file_metadata):
        op = S3KeySensor(
            task_id="s3_key_sensor",
            bucket_key=["s3://test_bucket/file*", "s3://test_bucket/*.zip"],
            wildcard_match=True,
        )

        mock_get_file_metadata.side_effect = [[{"Key": "file1", "Size": 0}], []]
        assert op.poke(None) is False

        mock_get_file_metadata.side_effect = [[{"Key": "file1", "Size": 0}], [{"Key": "file2", "Size": 0}]]
        assert op.poke(None) is False

        mock_get_file_metadata.side_effect = [[{"Key": "file1", "Size": 0}], [{"Key": "test.zip", "Size": 0}]]
        assert op.poke(None) is True

        mock_get_file_metadata.assert_any_call("file", "test_bucket")
        mock_get_file_metadata.assert_any_call("", "test_bucket")

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    def test_poke_with_check_function(self, mock_head_object):
        def check_fn(files: list) -> bool:
            return all(f.get("Size", 0) > 0 for f in files)

        op = S3KeySensor(task_id="s3_key_sensor", bucket_key="s3://test_bucket/file", check_fn=check_fn)

        mock_head_object.return_value = {"ContentLength": 0}
        assert op.poke(None) is False

        mock_head_object.return_value = {"ContentLength": 1}
        assert op.poke(None) is True

    @pytest.mark.parametrize(
        "key, pattern, expected",
        [
            ("test.csv", r"[a-z]+\.csv", True),
            ("test.txt", r"test/[a-z]+\.csv", False),
            ("test/test.csv", r"test/[a-z]+\.csv", True),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.get_file_metadata")
    def test_poke_with_use_regex(self, mock_get_file_metadata, key, pattern, expected):
        op = S3KeySensor(
            task_id="s3_key_sensor_async",
            bucket_key=pattern,
            bucket_name="test_bucket",
            use_regex=True,
        )
        mock_get_file_metadata.return_value = [{"Key": key, "Size": 0}]
        assert op.poke(None) is expected

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3KeySensor.poke", return_value=False)
    def test_s3_key_sensor_execute_complete_success_with_keys(self, mock_poke):
        """
        Asserts that a task is completed with success status and check function
        """

        def check_fn(files: list) -> bool:
            return all(f.get("Size", 0) > 0 for f in files)

        sensor = S3KeySensor(
            task_id="s3_key_sensor_async",
            bucket_key="key",
            bucket_name="bucket",
            check_fn=check_fn,
            deferrable=True,
        )
        assert (
            sensor.execute_complete(context={}, event={"status": "running", "files": [{"Size": 10}]}) is None
        )

    def test_fail_execute_complete(self):
        op = S3KeySensor(
            task_id="s3_key_sensor",
            bucket_key=["s3://test_bucket/file*", "s3://test_bucket/*.zip"],
            wildcard_match=True,
        )
        message = "error"
        with pytest.raises(AirflowException, match=message):
            op.execute_complete(context={}, event={"status": "error", "message": message})

    @pytest.mark.parametrize(
        "metadata_keys, expected",
        [
            (["Size", "Key"], True),
            (["Content"], False),
            (None, True),
        ],
    )
    @mock_aws
    def test_custom_metadata_default_return_vals(self, metadata_keys, expected):
        def check_fn(files: list) -> bool:
            for f in files:
                if "Size" not in f or "Key" not in f:
                    return False
            return True

        hook = S3Hook()
        hook.create_bucket(bucket_name="test-bucket")
        hook.load_string(
            bucket_name="test-bucket",
            key="test-key",
            string_data="test-body",
        )
        op = S3KeySensor(
            task_id="test-metadata",
            bucket_key="test-key",
            bucket_name="test-bucket",
            metadata_keys=metadata_keys,
            check_fn=check_fn,
        )
        assert op.poke(None) is expected

    @mock_aws
    def test_custom_metadata_default_custom_vals(self):
        def check_fn(files: list) -> bool:
            for f in files:
                if "LastModified" not in f or "ETag" not in f or "Size" in f:
                    return False
            return True

        hook = S3Hook()
        hook.create_bucket(bucket_name="test-bucket")
        hook.load_string(
            bucket_name="test-bucket",
            key="test-key",
            string_data="test-body",
        )

        op = S3KeySensor(
            task_id="test-metadata",
            bucket_key="test-key",
            bucket_name="test-bucket",
            metadata_keys=["LastModified", "ETag"],
            check_fn=check_fn,
        )
        assert op.poke(None) is True

    @mock_aws
    def test_custom_metadata_all_attributes(self):
        def check_fn(files: list) -> bool:
            hook = S3Hook()
            metadata_keys = set(hook.head_object(bucket_name="test-bucket", key="test-key").keys())
            test_data_keys = set(files[0].keys()) - {"Key"}

            return test_data_keys == metadata_keys

        hook = S3Hook()
        hook.create_bucket(bucket_name="test-bucket")
        hook.load_string(
            bucket_name="test-bucket",
            key="test-key",
            string_data="test-body",
        )

        op = S3KeySensor(
            task_id="test-metadata",
            bucket_key="test-key",
            bucket_name="test-bucket",
            metadata_keys=["*"],
            check_fn=check_fn,
        )
        assert op.poke(None) is True

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.get_file_metadata")
    def test_custom_metadata_wildcard(self, mock_file_metadata, mock_head_object):
        def check_fn(files: list) -> bool:
            for f in files:
                if "ETag" not in f or "MissingMeta" not in f:
                    return False
            return True

        op = S3KeySensor(
            task_id="test-head-metadata",
            bucket_key=["s3://test-bucket/test-key*"],
            metadata_keys=["MissingMeta", "ETag"],
            check_fn=check_fn,
            wildcard_match=True,
        )

        mock_file_metadata.return_value = [{"Key": "test-key", "ETag": 0}]
        mock_head_object.return_value = {"MissingMeta": 0, "ContentLength": 100}
        assert op.poke(None) is True
        mock_head_object.assert_called_once()

    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.head_object")
    @mock.patch("airflow.providers.amazon.aws.sensors.s3.S3Hook.get_file_metadata")
    def test_custom_metadata_wildcard_all_attributes(self, mock_file_metadata, mock_head_object):
        def check_fn(files: list) -> bool:
            for f in files:
                if "ContentLength" not in f or "MissingMeta" not in f:
                    return False
            return True

        op = S3KeySensor(
            task_id="test-head-metadata",
            bucket_key=["s3://test-bucket/test-key*"],
            metadata_keys=["*"],
            check_fn=check_fn,
            wildcard_match=True,
        )

        mock_file_metadata.return_value = [{"Key": "test-key", "ETag": 0}]
        mock_head_object.return_value = {"MissingMeta": 0, "ContentLength": 100}
        assert op.poke(None) is True
        mock_head_object.assert_called_once()

        mock_head_object.return_value = {"MissingMeta": 0}
        assert op.poke(None) is False


class TestS3KeysUnchangedSensor:
    def setup_method(self):
        self.dag = DAG(
            "unit_tests_aws_sensor_test_schedule_dag_once", start_date=DEFAULT_DATE, schedule="@once"
        )

        self.sensor = S3KeysUnchangedSensor(
            task_id="sensor_1",
            bucket_name="test-bucket",
            prefix="test-prefix/path",
            inactivity_period=12,
            poke_interval=0.1,
            min_objects=1,
            allow_delete=True,
            dag=self.dag,
        )

    def test_reschedule_mode_not_allowed(self):
        with pytest.raises(ValueError):
            S3KeysUnchangedSensor(
                task_id="sensor_2",
                bucket_name="test-bucket",
                prefix="test-prefix/path",
                poke_interval=0.1,
                mode="reschedule",
                dag=self.dag,
            )

    @pytest.mark.db_test
    def test_render_template_fields(self, clean_dags_and_dagruns):
        S3KeysUnchangedSensor(
            task_id="sensor_3",
            bucket_name="test-bucket",
            prefix="test-prefix/path",
            inactivity_period=12,
            poke_interval=0.1,
            min_objects=1,
            allow_delete=True,
            dag=self.dag,
        ).render_template_fields({})

    @time_machine.travel(DEFAULT_DATE)
    def test_files_deleted_between_pokes_throw_error(self):
        self.sensor.allow_delete = False
        self.sensor.is_keys_unchanged({"a", "b"})
        with pytest.raises(AirflowException):
            self.sensor.is_keys_unchanged({"a"})

    @pytest.mark.parametrize(
        "current_objects, expected_returns, inactivity_periods",
        [
            pytest.param(
                ({"a"}, {"a", "b"}, {"a", "b", "c"}),
                (False, False, False),
                (0, 0, 0),
                id="resetting inactivity period after key change",
            ),
            pytest.param(
                ({"a", "b"}, {"a"}, {"a", "c"}),
                (False, False, False),
                (0, 0, 0),
                id="item was deleted with option `allow_delete=True`",
            ),
            pytest.param(
                ({"a"}, {"a"}, {"a"}), (False, False, True), (0, 10, 20), id="inactivity period was exceeded"
            ),
            pytest.param(
                (set(), set(), set()), (False, False, False), (0, 10, 20), id="not pass if empty key is given"
            ),
        ],
    )
    def test_key_changes(self, current_objects, expected_returns, inactivity_periods, time_machine):
        time_machine.move_to(DEFAULT_DATE)
        for current, expected, period in zip(current_objects, expected_returns, inactivity_periods):
            assert self.sensor.is_keys_unchanged(current) == expected
            assert self.sensor.inactivity_seconds == period
            time_machine.coordinates.shift(10)

    def test_poke_succeeds_on_upload_complete(self, time_machine):
        time_machine.move_to(DEFAULT_DATE)
        self.sensor.hook = mock.MagicMock()
        self.sensor.hook.list_keys.return_value = {"a"}
        assert not self.sensor.poke(dict())
        time_machine.coordinates.shift(10)
        assert not self.sensor.poke(dict())
        time_machine.coordinates.shift(10)
        assert self.sensor.poke(dict())

    def test_fail_is_keys_unchanged(self):
        op = S3KeysUnchangedSensor(task_id="sensor", bucket_name="test-bucket", prefix="test-prefix/path")
        op.previous_objects = {"1", "2", "3"}
        current_objects = {"1", "2"}
        op.allow_delete = False
        message = "Illegal behavior: objects were deleted in"
        with pytest.raises(AirflowException, match=message):
            op.is_keys_unchanged(current_objects=current_objects)

    def test_fail_execute_complete(self):
        op = S3KeysUnchangedSensor(task_id="sensor", bucket_name="test-bucket", prefix="test-prefix/path")
        message = "test message"
        with pytest.raises(AirflowException, match=message):
            op.execute_complete(context={}, event={"status": "error", "message": message})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
AWS Batch service waiters.

.. seealso::

    - https://boto3.amazonaws.com/v1/documentation/api/latest/guide/clients.html#waiters
    - https://github.com/boto/botocore/blob/develop/botocore/waiter.py
"""

from __future__ import annotations

import json
import sys
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

import botocore.client
import botocore.exceptions
import botocore.waiter

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.batch_client import BatchClientHook

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.utils.task_log_fetcher import AwsTaskLogFetcher


class BatchWaitersHook(BatchClientHook):
    """
    A utility to manage waiters for AWS Batch services.

    .. code-block:: python

        import random
        from airflow.providers.amazon.aws.operators.batch_waiters import BatchWaiters

        # to inspect default waiters
        waiters = BatchWaiters()
        config = waiters.default_config  # type: Dict
        waiter_names = waiters.list_waiters()  # -> ["JobComplete", "JobExists", "JobRunning"]

        # The default_config is a useful stepping stone to creating custom waiters, e.g.
        custom_config = waiters.default_config  # this is a deepcopy
        # modify custom_config['waiters'] as necessary and get a new instance:
        waiters = BatchWaiters(waiter_config=custom_config)
        waiters.waiter_config  # check the custom configuration (this is a deepcopy)
        waiters.list_waiters()  # names of custom waiters

        # During the init for BatchWaiters, the waiter_config is used to build a waiter_model;
        # and note that this only occurs during the class init, to avoid any accidental mutations
        # of waiter_config leaking into the waiter_model.
        waiters.waiter_model  # -> botocore.waiter.WaiterModel object

        # The waiter_model is combined with the waiters.client to get a specific waiter
        # and the details of the config on that waiter can be further modified without any
        # accidental impact on the generation of new waiters from the defined waiter_model, e.g.
        waiters.get_waiter("JobExists").config.delay  # -> 5
        waiter = waiters.get_waiter("JobExists")  # -> botocore.waiter.Batch.Waiter.JobExists object
        waiter.config.delay = 10
        waiters.get_waiter("JobExists").config.delay  # -> 5 as defined by waiter_model

        # To use a specific waiter, update the config and call the `wait()` method for jobId, e.g.
        waiter = waiters.get_waiter("JobExists")  # -> botocore.waiter.Batch.Waiter.JobExists object
        waiter.config.delay = random.uniform(1, 10)  # seconds
        waiter.config.max_attempts = 10
        waiter.wait(jobs=[jobId])

    .. seealso::

        - https://www.2ndwatch.com/blog/use-waiters-boto3-write/
        - https://github.com/boto/botocore/blob/develop/botocore/waiter.py
        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#waiters
        - https://github.com/boto/botocore/tree/develop/botocore/data/ec2/2016-11-15
        - https://github.com/boto/botocore/issues/1915

    :param waiter_config:  a custom waiter configuration for AWS Batch services

    :param aws_conn_id: connection id of AWS credentials / region name. If None,
        credential boto3 strategy will be used
        (https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html).

    :param region_name: region name to use in AWS client.
        Override the AWS region in connection (if provided)
    """

    def __init__(self, *args, waiter_config: dict | None = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._default_config: dict | None = None
        self._waiter_config = waiter_config or self.default_config
        self._waiter_model = botocore.waiter.WaiterModel(self._waiter_config)

    @property
    def default_config(self) -> dict:
        """
        An immutable default waiter configuration.

        :return: a waiter configuration for AWS Batch services
        """
        if self._default_config is None:
            config_path = Path(__file__).with_name("batch_waiters.json").resolve()
            with open(config_path) as config_file:
                self._default_config = json.load(config_file)
        return deepcopy(self._default_config)  # avoid accidental mutation

    @property
    def waiter_config(self) -> dict:
        """
        An immutable waiter configuration for this instance; a ``deepcopy`` is returned by this property.

        During the init for BatchWaiters, the waiter_config is used to build a
        waiter_model and this only occurs during the class init, to avoid any
        accidental mutations of waiter_config leaking into the waiter_model.

        :return: a waiter configuration for AWS Batch services
        """
        return deepcopy(self._waiter_config)  # avoid accidental mutation

    @property
    def waiter_model(self) -> botocore.waiter.WaiterModel:
        """
        A configured waiter model used to generate waiters on AWS Batch services.

        :return: a waiter model for AWS Batch services
        """
        return self._waiter_model

    def get_waiter(
        self,
        waiter_name: str,
        parameters: dict[str, str] | None = None,
        config_overrides: dict[str, Any] | None = None,
        deferrable: bool = False,
        client=None,
    ) -> botocore.waiter.Waiter:
        """
        Get an AWS Batch service waiter, using the configured ``.waiter_model``.

        The ``.waiter_model`` is combined with the ``.client`` to get a specific waiter and
        the properties of that waiter can be modified without any accidental impact on the
        generation of new waiters from the ``.waiter_model``, e.g.

        .. code-block:: python

            waiters.get_waiter("JobExists").config.delay  # -> 5
            waiter = waiters.get_waiter("JobExists")  # a new waiter object
            waiter.config.delay = 10
            waiters.get_waiter("JobExists").config.delay  # -> 5 as defined by waiter_model

        To use a specific waiter, update the config and call the `wait()` method for jobId, e.g.

        .. code-block:: python

            import random

            waiter = waiters.get_waiter("JobExists")  # a new waiter object
            waiter.config.delay = random.uniform(1, 10)  # seconds
            waiter.config.max_attempts = 10
            waiter.wait(jobs=[jobId])

        :param waiter_name: The name of the waiter. The name should match
            the name (including the casing) of the key name in the waiter
            model file (typically this is CamelCasing); see ``.list_waiters``.

        :param parameters: unused, just here to match the method signature in base_aws
        :param config_overrides: unused, just here to match the method signature in base_aws
        :param deferrable: unused, just here to match the method signature in base_aws
        :param client: unused, just here to match the method signature in base_aws

        :return: a waiter object for the named AWS Batch service
        """
        return botocore.waiter.create_waiter_with_client(waiter_name, self.waiter_model, self.client)

    def list_waiters(self) -> list[str]:
        """
        List the waiters in a waiter configuration for AWS Batch services.

        :return: waiter names for AWS Batch services
        """
        return self.waiter_model.waiter_names

    def wait_for_job(
        self,
        job_id: str,
        delay: int | float | None = None,
        get_batch_log_fetcher: Callable[[str], AwsTaskLogFetcher | None] | None = None,
    ) -> None:
        """
        Wait for Batch job to complete.

        This assumes that the ``.waiter_model`` is configured using some
        variation of the ``.default_config`` so that it can generate waiters
        with the following names: "JobExists", "JobRunning" and "JobComplete".

        :param job_id: a Batch job ID

        :param delay:  A delay before polling for job status

        :param get_batch_log_fetcher: A method that returns batch_log_fetcher of
            type AwsTaskLogFetcher or None when the CloudWatch log stream hasn't been created yet.

        :raises: AirflowException

        .. note::
            This method adds a small random jitter to the ``delay`` (+/- 2 sec, >= 1 sec).
            Using a random interval helps to avoid AWS API throttle limits when many
            concurrent tasks request job-descriptions.

            It also modifies the ``max_attempts`` to use the ``sys.maxsize``,
            which allows Airflow to manage the timeout on waiting.
        """
        self.delay(delay)
        try:
            waiter = self.get_waiter("JobExists")
            waiter.config.delay = self.add_jitter(waiter.config.delay, width=2, minima=1)
            waiter.config.max_attempts = sys.maxsize  # timeout is managed by Airflow
            waiter.wait(jobs=[job_id])

            waiter = self.get_waiter("JobRunning")
            waiter.config.delay = self.add_jitter(waiter.config.delay, width=2, minima=1)
            waiter.config.max_attempts = sys.maxsize  # timeout is managed by Airflow
            waiter.wait(jobs=[job_id])

            batch_log_fetcher = None
            try:
                if get_batch_log_fetcher:
                    batch_log_fetcher = get_batch_log_fetcher(job_id)
                    if batch_log_fetcher:
                        batch_log_fetcher.start()
                waiter = self.get_waiter("JobComplete")
                waiter.config.delay = self.add_jitter(waiter.config.delay, width=2, minima=1)
                waiter.config.max_attempts = sys.maxsize  # timeout is managed by Airflow
                waiter.wait(jobs=[job_id])
            finally:
                if batch_log_fetcher:
                    batch_log_fetcher.stop()
                    batch_log_fetcher.join()

        except (botocore.exceptions.ClientError, botocore.exceptions.WaiterError) as err:
            raise AirflowException(err)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from collections.abc import Sequence

from airflow.providers.amazon.aws.utils.mixins import (
    AwsBaseHookMixin,
    AwsHookParams,
    AwsHookType,
    aws_template_fields,
)
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.types import NOTSET, ArgNotSet


class AwsBaseSensor(BaseSensorOperator, AwsBaseHookMixin[AwsHookType]):
    """
    Base AWS (Amazon) Sensor Class for build sensors in top of AWS Hooks.

    .. warning::
        Only for internal usage, this class might be changed, renamed or removed in the future
        without any further notice.

    Examples:
     .. code-block:: python

        from airflow.providers.amazon.aws.hooks.foo_bar import FooBarThinHook, FooBarThickHook


        class AwsFooBarSensor(AwsBaseSensor[FooBarThinHook]):
            aws_hook_class = FooBarThinHook

            def poke(self, context):
                pass


        class AwsFooBarSensor(AwsBaseSensor[FooBarThickHook]):
            aws_hook_class = FooBarThickHook

            def __init__(self, *, spam: str, **kwargs):
                super().__init__(**kwargs)
                self.spam = spam

            @property
            def _hook_parameters(self):
                return {**super()._hook_parameters, "spam": self.spam}

            def poke(self, context):
                pass

    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is None or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    :meta private:
    """

    template_fields: Sequence[str] = aws_template_fields()

    def __init__(
        self,
        *,
        aws_conn_id: str | None = "aws_default",
        region_name: str | None = None,
        verify: bool | str | None = None,
        botocore_config: dict | None = None,
        region: str | None | ArgNotSet = NOTSET,  # Required for `.partial` signature check
        **kwargs,
    ):
        additional_params = {} if region is NOTSET else {"region": region}
        hook_params = AwsHookParams.from_constructor(
            aws_conn_id, region_name, verify, botocore_config, additional_params=additional_params
        )
        super().__init__(**kwargs)
        self.aws_conn_id = hook_params.aws_conn_id
        self.region_name = hook_params.region_name
        self.verify = hook_params.verify
        self.botocore_config = hook_params.botocore_config
        self.validate_attributes()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook
from airflow.providers.amazon.aws.hooks.dms import DmsHook
from airflow.providers.amazon.aws.triggers.base import AwsBaseWaiterTrigger

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook


class DmsReplicationTerminalStatusTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when an AWS DMS Serverless replication is in a terminal state.

    :param replication_config_arn: The ARN of the replication config.
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        replication_config_arn: str,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 60,
        aws_conn_id: str | None = "aws_default",
    ) -> None:
        super().__init__(
            serialized_fields={"replication_config_arn": replication_config_arn},
            waiter_name="replication_terminal_status",
            waiter_delay=waiter_delay,
            waiter_args={"Filters": [{"Name": "replication-config-arn", "Values": [replication_config_arn]}]},
            waiter_max_attempts=waiter_max_attempts,
            failure_message="Replication failed to reach terminal status.",
            status_message="Status replication is",
            status_queries=["Replications[0].Status"],
            return_key="replication_config_arn",
            return_value=replication_config_arn,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return DmsHook(
            self.aws_conn_id,
            verify=self.verify,
            config=self.botocore_config,
        )


class DmsReplicationConfigDeletedTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when an AWS DMS Serverless replication config is deleted.

    :param replication_config_arn: The ARN of the replication config.
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        replication_config_arn: str,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 60,
        aws_conn_id: str | None = "aws_default",
    ) -> None:
        super().__init__(
            serialized_fields={"replication_config_arn": replication_config_arn},
            waiter_name="replication_config_deleted",
            waiter_delay=waiter_delay,
            waiter_args={"Filters": [{"Name": "replication-config-arn", "Values": [replication_config_arn]}]},
            waiter_max_attempts=waiter_max_attempts,
            failure_message="Replication config failed to be deleted.",
            status_message="Status replication config is",
            status_queries=["ReplicationConfigs[0].Status"],
            return_key="replication_config_arn",
            return_value=replication_config_arn,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return DmsHook(
            self.aws_conn_id,
            verify=self.verify,
            config=self.botocore_config,
        )


class DmsReplicationCompleteTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when an AWS DMS Serverless replication completes.

    :param replication_config_arn: The ARN of the replication config.
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        replication_config_arn: str,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 60,
        aws_conn_id: str | None = "aws_default",
    ) -> None:
        super().__init__(
            serialized_fields={"replication_config_arn": replication_config_arn},
            waiter_name="replication_complete",
            waiter_delay=waiter_delay,
            waiter_args={"Filters": [{"Name": "replication-config-arn", "Values": [replication_config_arn]}]},
            waiter_max_attempts=waiter_max_attempts,
            failure_message="Replication failed to complete.",
            status_message="Status replication is",
            status_queries=["Replications[0].Status"],
            return_key="replication_config_arn",
            return_value=replication_config_arn,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return DmsHook(
            self.aws_conn_id,
            verify=self.verify,
            config=self.botocore_config,
        )


class DmsReplicationStoppedTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when an AWS DMS Serverless replication is stopped.

    :param replication_config_arn: The ARN of the replication config.
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        replication_config_arn: str,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 60,
        aws_conn_id: str | None = "aws_default",
    ) -> None:
        super().__init__(
            serialized_fields={"replication_config_arn": replication_config_arn},
            waiter_name="replication_stopped",
            waiter_delay=waiter_delay,
            waiter_args={"Filters": [{"Name": "replication-config-arn", "Values": [replication_config_arn]}]},
            waiter_max_attempts=waiter_max_attempts,
            failure_message="Replication failed to stop.",
            status_message="Status replication is",
            status_queries=["Replications[0].Status"],
            return_key="replication_config_arn",
            return_value=replication_config_arn,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return DmsHook(
            self.aws_conn_id,
            verify=self.verify,
            config=self.botocore_config,
        )


class DmsReplicationDeprovisionedTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when an AWS DMS Serverless replication is de-provisioned.

    :param replication_config_arn: The ARN of the replication config.
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        replication_config_arn: str,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 60,
        aws_conn_id: str | None = "aws_default",
    ) -> None:
        super().__init__(
            serialized_fields={"replication_config_arn": replication_config_arn},
            waiter_name="replication_deprovisioned",
            waiter_delay=waiter_delay,
            waiter_args={"Filters": [{"Name": "replication-config-arn", "Values": [replication_config_arn]}]},
            waiter_max_attempts=waiter_max_attempts,
            failure_message="Replication failed to deprovision.",
            status_message="Status replication is",
            status_queries=["Replications[0].ProvisionData.ProvisionState"],
            return_key="replication_config_arn",
            return_value=replication_config_arn,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return DmsHook(
            self.aws_conn_id,
            verify=self.verify,
            config=self.botocore_config,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
import tempfile
from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.models import BaseOperator
from airflow.providers.amazon.aws.hooks.s3 import S3Hook

try:
    from airflow.providers.microsoft.azure.hooks.wasb import WasbHook
except ModuleNotFoundError as e:
    from airflow.exceptions import AirflowOptionalProviderFeatureException

    raise AirflowOptionalProviderFeatureException(e)

if TYPE_CHECKING:
    from airflow.utils.context import Context


class AzureBlobStorageToS3Operator(BaseOperator):
    """
    Operator transfers data from Azure Blob Storage to specified bucket in Amazon S3.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AzureBlobStorageToGCSOperator`

    :param wasb_conn_id: Reference to the wasb connection.
    :param container_name: Name of the container
    :param prefix: Prefix string which filters objects whose name begin with
        this prefix. (templated)
    :param delimiter: The delimiter by which you want to filter the objects. (templated)
        For e.g to lists the CSV files from in a directory in GCS you would use
        delimiter='.csv'.
    :param aws_conn_id: Connection id of the S3 connection to use
    :param dest_s3_key: The base S3 key to be used to store the files. (templated)
    :param dest_verify: Whether or not to verify SSL certificates for S3 connection.
        By default SSL certificates are verified.
        You can provide the following values:

        - ``False``: do not validate SSL certificates. SSL will still be used
                 (unless use_ssl is False), but SSL certificates will not be
                 verified.
        - ``path/to/cert/bundle.pem``: A filename of the CA cert bundle to uses.
                 You can specify this argument if you want to use a different
                 CA cert bundle than the one used by botocore.
    :param dest_s3_extra_args: Extra arguments that may be passed to the download/upload operations.
    :param replace: Whether or not to verify the existence of the files in the
        destination bucket.
        By default is set to False
        If set to True, will upload all the files replacing the existing ones in
        the destination bucket.
        If set to False, will upload only the files that are in the origin but not
        in the destination bucket.
    :param s3_acl_policy: Optional The string to specify the canned ACL policy for the
        object to be uploaded in S3
    :param wasb_extra_kargs: kwargs to pass to WasbHook
    :param s3_extra_kargs: kwargs to pass to S3Hook
    """

    template_fields: Sequence[str] = (
        "container_name",
        "prefix",
        "delimiter",
        "dest_s3_key",
    )

    def __init__(
        self,
        *,
        wasb_conn_id: str = "wasb_default",
        container_name: str,
        prefix: str | None = None,
        delimiter: str = "",
        aws_conn_id: str | None = "aws_default",
        dest_s3_key: str,
        dest_verify: str | bool | None = None,
        dest_s3_extra_args: dict | None = None,
        replace: bool = False,
        s3_acl_policy: str | None = None,
        wasb_extra_args: dict | None = None,
        s3_extra_args: dict | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.wasb_conn_id = wasb_conn_id
        self.container_name = container_name
        self.prefix = prefix
        self.delimiter = delimiter
        self.aws_conn_id = aws_conn_id
        self.dest_s3_key = dest_s3_key
        self.dest_verify = dest_verify
        self.dest_s3_extra_args = dest_s3_extra_args or {}
        self.replace = replace
        self.s3_acl_policy = s3_acl_policy
        self.wasb_extra_args = wasb_extra_args or {}
        self.s3_extra_args = s3_extra_args or {}

    def execute(self, context: Context) -> list[str]:
        # list all files in the Azure Blob Storage container
        wasb_hook = WasbHook(wasb_conn_id=self.wasb_conn_id, **self.wasb_extra_args)
        s3_hook = S3Hook(
            aws_conn_id=self.aws_conn_id,
            verify=self.dest_verify,
            extra_args=self.dest_s3_extra_args,
            **self.s3_extra_args,
        )

        self.log.info(
            "Getting list of the files in Container: %r; Prefix: %r; Delimiter: %r.",
            self.container_name,
            self.prefix,
            self.delimiter,
        )

        files = wasb_hook.get_blobs_list_recursive(
            container_name=self.container_name, prefix=self.prefix, endswith=self.delimiter
        )

        if not self.replace:
            # if we are not replacing -> list all files in the S3 bucket
            # and only keep those files which are present in
            # Azure Blob Storage and not in S3
            bucket_name, prefix = S3Hook.parse_s3_url(self.dest_s3_key)
            # look for the bucket and the prefix to avoid look into
            # parent directories/keys
            existing_files = s3_hook.list_keys(bucket_name, prefix=prefix)
            # in case that no files exists, return an empty array to avoid errors
            existing_files = existing_files or []
            # remove the prefix for the existing files to allow the match
            existing_files = [file.replace(f"{prefix}/", "", 1) for file in existing_files]
            files = list(set(files) - set(existing_files))

        if files:
            for file in files:
                with tempfile.NamedTemporaryFile() as temp_file:
                    dest_key = os.path.join(self.dest_s3_key, file)
                    self.log.info("Downloading data from blob: %s", file)
                    wasb_hook.get_file(
                        file_path=temp_file.name,
                        container_name=self.container_name,
                        blob_name=file,
                    )

                    self.log.info("Uploading data to s3: %s", dest_key)
                    s3_hook.load_file(
                        filename=temp_file.name,
                        key=dest_key,
                        replace=self.replace,
                        acl_policy=self.s3_acl_policy,
                    )
            self.log.info("All done, uploaded %d files to S3", len(files))
        else:
            self.log.info("All files are already in sync!")
        return files



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any
from unittest import mock

import pytest

from airflow.providers.amazon.aws.operators.glacier import (
    GlacierCreateJobOperator,
    GlacierUploadArchiveOperator,
)

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator

AWS_CONN_ID = "aws_default"
BUCKET_NAME = "airflow_bucket"
FILENAME = "path/to/file/"
GCP_CONN_ID = "google_cloud_default"
JOB_ID = "1a2b3c4d"
OBJECT_NAME = "file.csv"
TASK_ID = "glacier_job"
VAULT_NAME = "airflow"


class BaseGlacierOperatorsTests:
    op_class: type[AwsBaseOperator]
    default_op_kwargs: dict[str, Any]

    def test_base_aws_op_attributes(self):
        op = self.op_class(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = self.op_class(
            **self.default_op_kwargs,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42


class TestGlacierCreateJobOperator(BaseGlacierOperatorsTests):
    op_class = GlacierCreateJobOperator

    @pytest.fixture(autouse=True)
    def _setup_test_cases(self):
        self.default_op_kwargs = {"vault_name": VAULT_NAME, "task_id": TASK_ID}

    @mock.patch.object(GlacierCreateJobOperator, "hook", new_callable=mock.PropertyMock)
    def test_execute(self, hook_mock):
        op = self.op_class(aws_conn_id=None, **self.default_op_kwargs)
        op.execute(mock.MagicMock())
        hook_mock.return_value.retrieve_inventory.assert_called_once_with(vault_name=VAULT_NAME)

    def test_template_fields(self):
        operator = self.op_class(**self.default_op_kwargs)
        validate_template_fields(operator)


class TestGlacierUploadArchiveOperator(BaseGlacierOperatorsTests):
    op_class = GlacierUploadArchiveOperator

    @pytest.fixture(autouse=True)
    def _setup_test_cases(self):
        self.default_op_kwargs = {"vault_name": VAULT_NAME, "task_id": TASK_ID, "body": b"Test Data"}

    def test_execute(self):
        with mock.patch.object(self.op_class.aws_hook_class, "conn", new_callable=mock.PropertyMock) as m:
            op = self.op_class(aws_conn_id=None, **self.default_op_kwargs)
            op.execute(mock.MagicMock())
            m.return_value.upload_archive.assert_called_once_with(
                accountId=None,
                vaultName=VAULT_NAME,
                archiveDescription=None,
                body=b"Test Data",
                checksum=None,
            )

    def test_template_fields(self):
        operator = self.op_class(**self.default_op_kwargs)
        validate_template_fields(operator)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
from unittest import mock

import boto3
from moto import mock_aws

from airflow.models.dag import DAG
from airflow.providers.amazon.aws.transfers.http_to_s3 import HttpToS3Operator

EXAMPLE_URL = "http://www.example.com"


@mock.patch.dict("os.environ", AIRFLOW_CONN_HTTP_EXAMPLE=EXAMPLE_URL)
class TestHttpToS3Operator:
    def setup_method(self):
        args = {"owner": "airflow", "start_date": datetime.datetime(2017, 1, 1)}
        self.dag = DAG("test_dag_id", schedule=None, default_args=args)
        self.http_conn_id = "HTTP_EXAMPLE"
        self.response = b"Example.com fake response"
        self.endpoint = "/"
        self.s3_key = "test/test1.csv"
        self.s3_bucket = "dummy"

    def test_init(self):
        operator = HttpToS3Operator(
            task_id="http_to_s3_operator",
            http_conn_id=self.http_conn_id,
            endpoint=self.endpoint,
            s3_key=self.s3_key,
            s3_bucket=self.s3_bucket,
            dag=self.dag,
        )
        assert operator.endpoint == self.endpoint
        assert operator.s3_key == self.s3_key
        assert operator.s3_bucket == self.s3_bucket
        assert operator.http_conn_id == self.http_conn_id

    @mock_aws
    def test_execute(self, requests_mock):
        requests_mock.register_uri("GET", EXAMPLE_URL, content=self.response)
        conn = boto3.client("s3")
        conn.create_bucket(Bucket=self.s3_bucket)
        operator = HttpToS3Operator(
            task_id="s3_to_file_sensor",
            http_conn_id=self.http_conn_id,
            endpoint=self.endpoint,
            s3_key=self.s3_key,
            s3_bucket=self.s3_bucket,
            dag=self.dag,
        )
        operator.execute(None)

        objects_in_bucket = conn.list_objects(Bucket=self.s3_bucket, Prefix=self.s3_key)
        # there should be object found, and there should only be one object found
        assert len(objects_in_bucket["Contents"]) == 1
        # the object found should be consistent with dest_key specified earlier
        assert objects_in_bucket["Contents"][0]["Key"] == self.s3_key



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import inspect
import itertools
import time
from unittest import mock

import boto3
import pytest
from botocore.exceptions import ClientError, WaiterError
from botocore.waiter import SingleWaiterConfig, WaiterModel
from moto import mock_aws

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.batch_waiters import BatchWaitersHook
from airflow.providers.amazon.aws.utils.task_log_fetcher import AwsTaskLogFetcher

INTERMEDIATE_STATES = ("SUBMITTED", "PENDING", "RUNNABLE", "STARTING")
RUNNING_STATE = "RUNNING"
SUCCESS_STATE = "SUCCEEDED"
FAILED_STATE = "FAILED"
ALL_STATES = {*INTERMEDIATE_STATES, RUNNING_STATE, SUCCESS_STATE, FAILED_STATE}
AWS_REGION = "eu-west-1"


@pytest.fixture(scope="module")
def aws_region():
    return AWS_REGION


@mock_aws
@pytest.fixture
def patch_hook(monkeypatch, aws_region):
    """Patch hook object by dummy boto3 Batch client."""
    batch_client = boto3.client("batch", region_name=aws_region)
    monkeypatch.setattr(BatchWaitersHook, "conn", batch_client)


def test_batch_waiters(aws_region):
    assert inspect.isclass(BatchWaitersHook)
    batch_waiters = BatchWaitersHook(region_name=aws_region)
    assert isinstance(batch_waiters, BatchWaitersHook)


class TestBatchWaiters:
    @pytest.fixture(autouse=True)
    def setup_tests(self, patch_hook):
        self.job_id = "8ba9d676-4108-4474-9dca-8bbac1da9b19"
        self.region_name = AWS_REGION

        self.batch_waiters = BatchWaitersHook(region_name=self.region_name)
        assert self.batch_waiters.aws_conn_id == "aws_default"
        assert self.batch_waiters.region_name == self.region_name

        # don't pause in these unit tests
        self.mock_delay = mock.Mock(return_value=None)
        self.batch_waiters.delay = self.mock_delay
        self.mock_exponential_delay = mock.Mock(return_value=0)
        self.batch_waiters.exponential_delay = self.mock_exponential_delay

    def test_default_config(self):
        # the default config is used when no custom config is provided
        config = self.batch_waiters.default_config
        assert config == self.batch_waiters.waiter_config

        assert isinstance(config, dict)
        assert config["version"] == 2
        assert isinstance(config["waiters"], dict)

        waiters = list(sorted(config["waiters"].keys()))
        assert waiters == ["JobComplete", "JobExists", "JobRunning"]

    def test_list_waiters(self):
        # the default config is used when no custom config is provided
        config = self.batch_waiters.waiter_config

        assert isinstance(config["waiters"], dict)
        waiters = list(sorted(config["waiters"].keys()))
        assert waiters == ["JobComplete", "JobExists", "JobRunning"]
        assert waiters == self.batch_waiters.list_waiters()

    def test_waiter_model(self):
        model = self.batch_waiters.waiter_model
        assert isinstance(model, WaiterModel)

        # test some of the default config
        assert model.version == 2
        waiters = sorted(model.waiter_names)
        assert waiters == ["JobComplete", "JobExists", "JobRunning"]

        # test errors when requesting a waiter with the wrong name
        with pytest.raises(ValueError) as ctx:
            model.get_waiter("JobExist")
        assert "Waiter does not exist: JobExist" in str(ctx.value)

        # test some default waiter properties
        waiter = model.get_waiter("JobExists")
        assert isinstance(waiter, SingleWaiterConfig)
        assert waiter.max_attempts == 100
        waiter.max_attempts = 200
        assert waiter.max_attempts == 200
        assert waiter.delay == 2
        waiter.delay = 10
        assert waiter.delay == 10
        assert waiter.operation == "DescribeJobs"

    def test_wait_for_job(self):
        import sys

        # mock delay for speedy test
        mock_jitter = mock.Mock(return_value=0)
        self.batch_waiters.add_jitter = mock_jitter

        with mock.patch.object(self.batch_waiters, "get_waiter") as get_waiter:
            self.batch_waiters.wait_for_job(self.job_id)

            assert get_waiter.call_args_list == [
                mock.call("JobExists"),
                mock.call("JobRunning"),
                mock.call("JobComplete"),
            ]

            mock_waiter = get_waiter.return_value
            mock_waiter.wait.assert_called_with(jobs=[self.job_id])
            assert mock_waiter.wait.call_count == 3

            mock_config = mock_waiter.config
            assert mock_config.delay == 0
            assert mock_config.max_attempts == sys.maxsize

    def test_wait_for_job_with_cloudwatch_logs(self):
        # mock delay for speedy test
        mock_jitter = mock.Mock(return_value=0)
        self.batch_waiters.add_jitter = mock_jitter

        batch_log_fetcher = mock.Mock(spec=AwsTaskLogFetcher)
        mock_get_batch_log_fetcher = mock.Mock(return_value=batch_log_fetcher)

        thread_start = mock.Mock(side_effect=lambda: time.sleep(2))
        thread_stop = mock.Mock(side_effect=lambda: time.sleep(2))
        thread_join = mock.Mock(side_effect=lambda: time.sleep(2))

        with (
            mock.patch.object(self.batch_waiters, "get_waiter") as mock_get_waiter,
            mock.patch.object(batch_log_fetcher, "start", thread_start) as mock_fetcher_start,
            mock.patch.object(batch_log_fetcher, "stop", thread_stop) as mock_fetcher_stop,
            mock.patch.object(batch_log_fetcher, "join", thread_join) as mock_fetcher_join,
        ):
            # Run the wait_for_job method
            self.batch_waiters.wait_for_job(self.job_id, get_batch_log_fetcher=mock_get_batch_log_fetcher)

            # Assertions
            assert mock_get_waiter.call_args_list == [
                mock.call("JobExists"),
                mock.call("JobRunning"),
                mock.call("JobComplete"),
            ]

            mock_get_waiter.return_value.wait.assert_called_with(jobs=[self.job_id])
            mock_get_batch_log_fetcher.assert_called_with(self.job_id)
            mock_fetcher_start.assert_called_once()
            mock_fetcher_stop.assert_called_once()
            mock_fetcher_join.assert_called_once()

    def test_wait_for_job_raises_for_client_error(self):
        # mock delay for speedy test
        mock_jitter = mock.Mock(return_value=0)
        self.batch_waiters.add_jitter = mock_jitter

        with mock.patch.object(self.batch_waiters, "get_waiter") as get_waiter:
            mock_waiter = get_waiter.return_value
            mock_waiter.wait.side_effect = ClientError(
                error_response={"Error": {"Code": "TooManyRequestsException"}},
                operation_name="get job description",
            )
            with pytest.raises(AirflowException):
                self.batch_waiters.wait_for_job(self.job_id)

            assert get_waiter.call_args_list == [mock.call("JobExists")]
            mock_waiter.wait.assert_called_with(jobs=[self.job_id])
            assert mock_waiter.wait.call_count == 1

    def test_wait_for_job_raises_for_waiter_error(self):
        # mock delay for speedy test
        mock_jitter = mock.Mock(return_value=0)
        self.batch_waiters.add_jitter = mock_jitter

        with mock.patch.object(self.batch_waiters, "get_waiter") as get_waiter:
            mock_waiter = get_waiter.return_value
            mock_waiter.wait.side_effect = WaiterError(
                name="JobExists", reason="unit test error", last_response={}
            )
            with pytest.raises(AirflowException):
                self.batch_waiters.wait_for_job(self.job_id)

            assert get_waiter.call_args_list == [mock.call("JobExists")]
            mock_waiter.wait.assert_called_with(jobs=[self.job_id])
            assert mock_waiter.wait.call_count == 1


class TestBatchJobWaiters:
    """Test default waiters."""

    @pytest.fixture(autouse=True)
    def setup_tests(self, patch_hook):
        """Mock `describe_jobs` method before each test run."""
        self.batch_waiters = BatchWaitersHook(region_name=AWS_REGION)
        self.client = self.batch_waiters.client

        with mock.patch.object(self.client, "describe_jobs") as m:
            self.mock_describe_jobs = m
            yield

    @staticmethod
    def describe_jobs_response(job_id: str = "mock-job-id", status: str = INTERMEDIATE_STATES[0]):
        """
        Helper function for generate minimal DescribeJobs response for single job.
        https://docs.aws.amazon.com/batch/latest/APIReference/API_DescribeJobs.html
        """
        assert job_id
        assert status in ALL_STATES

        return {"jobs": [{"jobId": job_id, "status": status}]}

    @pytest.mark.parametrize("status", sorted(ALL_STATES))
    def test_job_exists_waiter_exists(self, status: str):
        """Test `JobExists` when response return dictionary regardless state."""
        self.mock_describe_jobs.return_value = self.describe_jobs_response(
            job_id="job-exist-success", status=status
        )
        job_exists_waiter = self.batch_waiters.get_waiter("JobExists")
        job_exists_waiter.config.delay = 0.01
        job_exists_waiter.config.max_attempts = 5
        job_exists_waiter.wait(jobs=["job-exist-success"])
        assert self.mock_describe_jobs.called

    def test_job_exists_waiter_missing(self):
        """Test `JobExists` waiter when response return empty dictionary."""
        self.mock_describe_jobs.return_value = {"jobs": []}

        job_exists_waiter = self.batch_waiters.get_waiter("JobExists")
        job_exists_waiter.config.delay = 0.01
        job_exists_waiter.config.max_attempts = 20
        with pytest.raises(WaiterError, match="Waiter JobExists failed"):
            job_exists_waiter.wait(jobs=["job-missing"])
        assert self.mock_describe_jobs.called

    @pytest.mark.parametrize("status", [RUNNING_STATE, SUCCESS_STATE, FAILED_STATE])
    def test_job_running_waiter_change_to_waited_state(self, status):
        """Test `JobRunning` waiter reach expected state."""
        job_id = "job-running"
        self.mock_describe_jobs.side_effect = [
            # Emulate change job status before one of expected states.
            # SUBMITTED -> PENDING -> RUNNABLE -> STARTING
            *itertools.chain.from_iterable(
                itertools.repeat(self.describe_jobs_response(job_id=job_id, status=inter_status), 3)
                for inter_status in INTERMEDIATE_STATES
            ),
            # Expected status
            self.describe_jobs_response(job_id=job_id, status=status),
            RuntimeError("This should not raise"),
        ]

        job_running_waiter = self.batch_waiters.get_waiter("JobRunning")
        job_running_waiter.config.delay = 0.01
        job_running_waiter.config.max_attempts = 20
        job_running_waiter.wait(jobs=[job_id])
        assert self.mock_describe_jobs.called

    @pytest.mark.parametrize("status", INTERMEDIATE_STATES)
    def test_job_running_waiter_max_attempt_exceeded(self, status):
        """Test `JobRunning` waiter run out of attempts."""
        job_id = "job-running-inf"
        self.mock_describe_jobs.side_effect = itertools.repeat(
            self.describe_jobs_response(job_id=job_id, status=status)
        )
        job_running_waiter = self.batch_waiters.get_waiter("JobRunning")
        job_running_waiter.config.delay = 0.01
        job_running_waiter.config.max_attempts = 20
        with pytest.raises(WaiterError, match="Waiter JobRunning failed: Max attempts exceeded"):
            job_running_waiter.wait(jobs=[job_id])
        assert self.mock_describe_jobs.called

    def test_job_complete_waiter_succeeded(self):
        """Test `JobComplete` waiter reach `SUCCEEDED` status."""
        job_id = "job-succeeded"
        self.mock_describe_jobs.side_effect = [
            *itertools.repeat(self.describe_jobs_response(job_id=job_id, status=RUNNING_STATE), 10),
            self.describe_jobs_response(job_id=job_id, status=SUCCESS_STATE),
            RuntimeError("This should not raise"),
        ]

        job_complete_waiter = self.batch_waiters.get_waiter("JobComplete")
        job_complete_waiter.config.delay = 0.01
        job_complete_waiter.config.max_attempts = 20
        job_complete_waiter.wait(jobs=[job_id])
        assert self.mock_describe_jobs.called

    def test_job_complete_waiter_failed(self):
        """Test `JobComplete` waiter reach `FAILED` status."""
        job_id = "job-failed"
        self.mock_describe_jobs.side_effect = [
            *itertools.repeat(self.describe_jobs_response(job_id=job_id, status=RUNNING_STATE), 10),
            self.describe_jobs_response(job_id=job_id, status=FAILED_STATE),
            RuntimeError("This should not raise"),
        ]

        job_complete_waiter = self.batch_waiters.get_waiter("JobComplete")
        job_complete_waiter.config.delay = 0.01
        job_complete_waiter.config.max_attempts = 20
        with pytest.raises(
            WaiterError, match="Waiter JobComplete failed: Waiter encountered a terminal failure state"
        ):
            job_complete_waiter.wait(jobs=[job_id])
        assert self.mock_describe_jobs.called

    def test_job_complete_waiter_max_attempt_exceeded(self):
        """Test `JobComplete` waiter run out of attempts."""
        job_id = "job-running-inf"
        self.mock_describe_jobs.side_effect = itertools.repeat(
            self.describe_jobs_response(job_id=job_id, status=RUNNING_STATE)
        )
        job_running_waiter = self.batch_waiters.get_waiter("JobComplete")
        job_running_waiter.config.delay = 0.01
        job_running_waiter.config.max_attempts = 20
        with pytest.raises(WaiterError, match="Waiter JobComplete failed: Max attempts exceeded"):
            job_running_waiter.wait(jobs=[job_id])
        assert self.mock_describe_jobs.called



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import abstractmethod
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.utils.waiter_with_logging import async_wait
from airflow.triggers.base import BaseTrigger, TriggerEvent
from airflow.utils.helpers import prune_dict

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook


class AwsBaseWaiterTrigger(BaseTrigger):
    """
    Base class for all AWS Triggers that follow the "standard" model of just waiting on a waiter.

    Subclasses need to implement the hook() method.

    :param serialized_fields: Fields that are specific to the subclass trigger and need to be serialized
        to be passed to the __init__ method on deserialization.
        The conn id, region, and waiter delay & attempts are always serialized.
        format: {<parameter_name>: <parameter_value>}

    :param waiter_name: The name of the (possibly custom) boto waiter to use.

    :param waiter_args: The arguments to pass to the waiter.
    :param failure_message: The message to log if a failure state is reached.
    :param status_message: The message logged when printing the status of the service.
    :param status_queries: A list containing the JMESPath queries to retrieve status information from
        the waiter response. See https://jmespath.org/tutorial.html

    :param return_key: The key to use for the return_value in the TriggerEvent this emits on success.
        Defaults to "value".
    :param return_value: A value that'll be returned in the return_key field of the TriggerEvent.
        Set to None if there is nothing to return.

    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param waiter_config_overrides: A dict to update waiter's default configuration. Only specified keys will
        be updated.
    :param aws_conn_id: The Airflow connection used for AWS credentials. To be used to build the hook.
    :param region_name: The AWS region where the resources to watch are. To be used to build the hook.
    :param verify: Whether or not to verify SSL certificates. To be used to build the hook.
        See: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client.
        To be used to build the hook. For available key-values see:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    def __init__(
        self,
        *,
        serialized_fields: dict[str, Any],
        waiter_name: str,
        waiter_args: dict[str, Any],
        failure_message: str,
        status_message: str,
        status_queries: list[str],
        return_key: str = "value",
        return_value: Any,
        waiter_delay: int,
        waiter_max_attempts: int,
        waiter_config_overrides: dict[str, Any] | None = None,
        aws_conn_id: str | None,
        region_name: str | None = None,
        verify: bool | str | None = None,
        botocore_config: dict | None = None,
    ):
        super().__init__()
        # parameters that should be hardcoded in the child's implem
        self.serialized_fields = serialized_fields

        self.waiter_name = waiter_name
        self.waiter_args = waiter_args
        self.failure_message = failure_message
        self.status_message = status_message
        self.status_queries = status_queries
        self.waiter_config_overrides = waiter_config_overrides

        self.return_key = return_key
        self.return_value = return_value

        # parameters that should be passed directly from the child's parameters
        self.waiter_delay = waiter_delay
        self.attempts = waiter_max_attempts
        self.aws_conn_id = aws_conn_id
        self.region_name = region_name
        self.verify = verify
        self.botocore_config = botocore_config

    def serialize(self) -> tuple[str, dict[str, Any]]:
        # here we put together the "common" params,
        # and whatever extras we got from the subclass in serialized_fields
        params = dict(
            {
                "waiter_delay": self.waiter_delay,
                "waiter_max_attempts": self.attempts,
                "aws_conn_id": self.aws_conn_id,
            },
            **self.serialized_fields,
        )

        # if we serialize the None value from this, it breaks subclasses that don't have it in their ctor.
        params.update(
            prune_dict(
                {
                    # Keep previous behaviour when empty string in region_name evaluated as `None`
                    "region_name": self.region_name or None,
                    "verify": self.verify,
                    "botocore_config": self.botocore_config,
                }
            )
        )

        return (
            # remember that self is an instance of the subclass here, not of this class.
            self.__class__.__module__ + "." + self.__class__.__qualname__,
            params,
        )

    @abstractmethod
    def hook(self) -> AwsGenericHook:
        """Override in subclasses to return the right hook."""

    async def run(self) -> AsyncIterator[TriggerEvent]:
        hook = self.hook()
        async with await hook.get_async_conn() as client:
            waiter = hook.get_waiter(
                self.waiter_name,
                deferrable=True,
                client=client,
                config_overrides=self.waiter_config_overrides,
            )
            await async_wait(
                waiter,
                self.waiter_delay,
                self.attempts,
                self.waiter_args,
                self.failure_message,
                self.status_message,
                self.status_queries,
            )
            yield TriggerEvent({"status": "success", self.return_key: self.return_value})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains AWS SQS hook."""

from __future__ import annotations

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook


class SqsHook(AwsBaseHook):
    """
    Interact with Amazon Simple Queue Service.

    Provide thin wrapper around :external+boto3:py:class:`boto3.client("sqs") <SQS.Client>`.

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs["client_type"] = "sqs"
        super().__init__(*args, **kwargs)

    def create_queue(self, queue_name: str, attributes: dict | None = None) -> dict:
        """
        Create queue using connection object.

        .. seealso::
            - :external+boto3:py:meth:`SQS.Client.create_queue`

        :param queue_name: name of the queue.
        :param attributes: additional attributes for the queue (default: None)
        :return: dict with the information about the queue.
        """
        return self.get_conn().create_queue(QueueName=queue_name, Attributes=attributes or {})

    def send_message(
        self,
        queue_url: str,
        message_body: str,
        delay_seconds: int = 0,
        message_attributes: dict | None = None,
        message_group_id: str | None = None,
        message_deduplication_id: str | None = None,
    ) -> dict:
        """
        Send message to the queue.

        .. seealso::
            - :external+boto3:py:meth:`SQS.Client.send_message`

        :param queue_url: queue url
        :param message_body: the contents of the message
        :param delay_seconds: seconds to delay the message
        :param message_attributes: additional attributes for the message (default: None)
        :param message_group_id: This applies only to FIFO (first-in-first-out) queues. (default: None)
        :param message_deduplication_id: This applies only to FIFO (first-in-first-out) queues.
        :return: dict with the information about the message sent
        """
        params = {
            "QueueUrl": queue_url,
            "MessageBody": message_body,
            "DelaySeconds": delay_seconds,
            "MessageAttributes": message_attributes or {},
        }
        if message_group_id:
            params["MessageGroupId"] = message_group_id
        if message_deduplication_id:
            params["MessageDeduplicationId"] = message_deduplication_id

        return self.get_conn().send_message(**params)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.utils.emailer import send_email


class TestSendEmailSes:
    @mock.patch("airflow.providers.amazon.aws.utils.emailer.SesHook")
    def test_send_ses_email(self, mock_hook):
        send_email(
            to="to@test.com",
            subject="subject",
            html_content="content",
            from_email="From Test <from@test.com>",
            custom_headers={"X-Test-Header": "test-val"},
        )

        mock_hook.return_value.send_email.assert_called_once_with(
            mail_from="From Test <from@test.com>",
            to="to@test.com",
            subject="subject",
            html_content="content",
            bcc=None,
            cc=None,
            files=None,
            mime_charset="utf-8",
            mime_subtype="mixed",
            custom_headers={"X-Test-Header": "test-val"},
        )

    @mock.patch("airflow.providers.amazon.aws.utils.emailer.SesHook")
    def test_send_ses_email_no_from_mail(self, mock_hook):
        with pytest.raises(
            RuntimeError, match="The `from_email' configuration has to be set for the SES emailer."
        ):
            send_email(to="to@test.com", subject="subject", html_content="content")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING
from unittest import mock

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.comprehend import ComprehendHook
from airflow.providers.amazon.aws.operators.comprehend import (
    ComprehendBaseOperator,
    ComprehendCreateDocumentClassifierOperator,
    ComprehendStartPiiEntitiesDetectionJobOperator,
)
from airflow.utils.types import NOTSET

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import BaseAwsConnection

INPUT_DATA_CONFIG = {
    "S3Uri": "s3://input-data-comprehend/sample_data.txt",
    "InputFormat": "ONE_DOC_PER_LINE",
}
OUTPUT_DATA_CONFIG = {"S3Uri": "s3://output-data-comprehend/redacted_output/"}
LANGUAGE_CODE = "en"
ROLE_ARN = "role_arn"


class TestComprehendBaseOperator:
    @pytest.mark.parametrize("aws_conn_id", [None, NOTSET, "aws_test_conn"])
    @pytest.mark.parametrize("region_name", [None, NOTSET, "ca-central-1"])
    def test_initialize_comprehend_base_operator(self, aws_conn_id, region_name):
        op_kw = {"aws_conn_id": aws_conn_id, "region_name": region_name}
        op_kw = {k: v for k, v in op_kw.items() if v is not NOTSET}

        comprehend_base_op = ComprehendBaseOperator(
            task_id="comprehend_base_operator",
            input_data_config=INPUT_DATA_CONFIG,
            output_data_config=OUTPUT_DATA_CONFIG,
            language_code=LANGUAGE_CODE,
            data_access_role_arn=ROLE_ARN,
            **op_kw,
        )

        assert comprehend_base_op.aws_conn_id == (aws_conn_id if aws_conn_id is not NOTSET else "aws_default")
        assert comprehend_base_op.region_name == (region_name if region_name is not NOTSET else None)

    @mock.patch.object(ComprehendBaseOperator, "hook", new_callable=mock.PropertyMock)
    def test_initialize_comprehend_base_operator_hook(self, comprehend_base_operator_mock_hook):
        comprehend_base_op = ComprehendBaseOperator(
            task_id="comprehend_base_operator",
            input_data_config=INPUT_DATA_CONFIG,
            output_data_config=OUTPUT_DATA_CONFIG,
            language_code=LANGUAGE_CODE,
            data_access_role_arn=ROLE_ARN,
        )
        mocked_hook = mock.MagicMock(name="MockHook")
        mocked_client = mock.MagicMock(name="MockClient")
        mocked_hook.conn = mocked_client
        comprehend_base_operator_mock_hook.return_value = mocked_hook
        assert comprehend_base_op.client == mocked_client
        comprehend_base_operator_mock_hook.assert_called_once()


class TestComprehendStartPiiEntitiesDetectionJobOperator:
    JOB_ID = "random-job-id-1234567"
    MODE = "ONLY_REDACTION"
    JOB_NAME = "TEST_START_PII_ENTITIES_DETECTION_JOB-1"
    DEFAULT_JOB_NAME_STARTS_WITH = "start_pii_entities_detection_job"
    REDACTION_CONFIG = {"PiiEntityTypes": ["NAME", "ADDRESS"], "MaskMode": "REPLACE_WITH_PII_ENTITY_TYPE"}

    @pytest.fixture
    def mock_conn(self) -> Generator[BaseAwsConnection, None, None]:
        with mock.patch.object(ComprehendHook, "conn") as _conn:
            _conn.start_pii_entities_detection_job.return_value = {"JobId": self.JOB_ID}
            yield _conn

    @pytest.fixture
    def comprehend_hook(self) -> Generator[ComprehendHook, None, None]:
        with mock_aws():
            hook = ComprehendHook(aws_conn_id="aws_default")
            yield hook

    def setup_method(self):
        self.operator = ComprehendStartPiiEntitiesDetectionJobOperator(
            task_id="start_pii_entities_detection_job",
            input_data_config=INPUT_DATA_CONFIG,
            output_data_config=OUTPUT_DATA_CONFIG,
            data_access_role_arn=ROLE_ARN,
            mode=self.MODE,
            language_code=LANGUAGE_CODE,
            start_pii_entities_kwargs={"JobName": self.JOB_NAME, "RedactionConfig": self.REDACTION_CONFIG},
        )
        self.operator.defer = mock.MagicMock()

    def test_init(self):
        assert self.operator.input_data_config == INPUT_DATA_CONFIG
        assert self.operator.output_data_config == OUTPUT_DATA_CONFIG
        assert self.operator.data_access_role_arn == ROLE_ARN
        assert self.operator.mode == self.MODE
        assert self.operator.language_code == LANGUAGE_CODE
        assert self.operator.start_pii_entities_kwargs.get("JobName") == self.JOB_NAME
        assert self.operator.start_pii_entities_kwargs.get("RedactionConfig") == self.REDACTION_CONFIG

    @mock.patch.object(ComprehendHook, "conn")
    def test_start_pii_entities_detection_job_name_starts_with_service_name(self, comprehend_mock_conn):
        self.op = ComprehendStartPiiEntitiesDetectionJobOperator(
            task_id="start_pii_entities_detection_job",
            input_data_config=INPUT_DATA_CONFIG,
            output_data_config=OUTPUT_DATA_CONFIG,
            data_access_role_arn=ROLE_ARN,
            mode=self.MODE,
            language_code=LANGUAGE_CODE,
            start_pii_entities_kwargs={"RedactionConfig": self.REDACTION_CONFIG},
        )
        self.op.wait_for_completion = False
        self.op.execute({})
        assert self.op.start_pii_entities_kwargs.get("JobName").startswith(self.DEFAULT_JOB_NAME_STARTS_WITH)
        comprehend_mock_conn.start_pii_entities_detection_job.assert_called_once_with(
            InputDataConfig=INPUT_DATA_CONFIG,
            OutputDataConfig=OUTPUT_DATA_CONFIG,
            Mode=self.MODE,
            DataAccessRoleArn=ROLE_ARN,
            LanguageCode=LANGUAGE_CODE,
            RedactionConfig=self.REDACTION_CONFIG,
            JobName=self.op.start_pii_entities_kwargs.get("JobName"),
        )

    @pytest.mark.parametrize(
        "wait_for_completion, deferrable",
        [
            pytest.param(False, False, id="no_wait"),
            pytest.param(True, False, id="wait"),
            pytest.param(False, True, id="defer"),
        ],
    )
    @mock.patch.object(ComprehendHook, "get_waiter")
    def test_start_pii_entities_detection_job_wait_combinations(
        self, _, wait_for_completion, deferrable, mock_conn, comprehend_hook
    ):
        self.operator.wait_for_completion = wait_for_completion
        self.operator.deferrable = deferrable

        response = self.operator.execute({})

        assert response == self.JOB_ID
        assert comprehend_hook.get_waiter.call_count == wait_for_completion
        assert self.operator.defer.call_count == deferrable

    def test_template_fields(self):
        validate_template_fields(self.operator)


class TestComprehendCreateDocumentClassifierOperator:
    CLASSIFIER_ARN = (
        "arn:aws:comprehend:us-east-1:123456789012:document-classifier/insurance-classifier/version/v1"
    )
    ROLE_ARN = "arn:aws:iam::123456789012:role/ComprehendExecutionRole"
    INPUT_DATA_CONFIG = {
        "DataFormat": "COMPREHEND_CSV",
        "S3Uri": "s3://test/native-doc.csv",
        "DocumentType": "SEMI_STRUCTURED_DOCUMENT",
        "Documents": {"S3Uri": "s3://test/input-docs/"},
        "DocumentReaderConfig": {
            "DocumentReadAction": "TEXTRACT_DETECT_DOCUMENT_TEXT",
            "DocumentReadMode": "SERVICE_DEFAULT",
        },
    }

    @pytest.fixture
    def mock_conn(self) -> Generator[BaseAwsConnection, None, None]:
        with mock.patch.object(ComprehendHook, "conn") as _conn:
            _conn.create_document_classifier.return_value = {"DocumentClassifierArn": self.CLASSIFIER_ARN}
            yield _conn

    @pytest.fixture
    def comprehend_hook(self) -> Generator[ComprehendHook, None, None]:
        with mock_aws():
            hook = ComprehendHook(aws_conn_id="aws_default")
            yield hook

    def setup_method(self):
        self.operator = ComprehendCreateDocumentClassifierOperator(
            task_id="create_document_classifier",
            data_access_role_arn=self.ROLE_ARN,
            document_classifier_name="custom_test_document_classifier",
            input_data_config=self.INPUT_DATA_CONFIG,
            output_data_config={"S3Uri": "s3://test/training_output/"},
            language_code="en",
            mode="MULTI_CLASS",
            document_classifier_kwargs={"VersionName": "v1"},
        )
        self.operator.defer = mock.MagicMock()

    def test_init(self):
        assert self.operator.input_data_config == self.INPUT_DATA_CONFIG
        assert self.operator.data_access_role_arn == self.ROLE_ARN
        assert self.operator.document_classifier_name == "custom_test_document_classifier"
        assert self.operator.output_data_config == {"S3Uri": "s3://test/training_output/"}
        assert self.operator.mode == "MULTI_CLASS"
        assert self.operator.language_code == "en"
        assert self.operator.document_classifier_kwargs == {"VersionName": "v1"}
        assert self.operator.fail_on_warnings is False

    @mock.patch.object(ComprehendHook, "conn")
    def test_create_document_classifier(self, mock_conn):
        self.op = ComprehendCreateDocumentClassifierOperator(
            task_id="create_document_classifier",
            data_access_role_arn=self.ROLE_ARN,
            document_classifier_name="custom_test_document_classifier",
            input_data_config=self.INPUT_DATA_CONFIG,
            output_data_config={"S3Uri": "s3://test/training_output/"},
            language_code="en",
            mode="MULTI_CLASS",
            document_classifier_kwargs={"VersionName": "v1"},
        )
        self.op.wait_for_completion = False
        self.op.execute({})
        mock_conn.create_document_classifier.assert_called_once_with(
            DocumentClassifierName="custom_test_document_classifier",
            DataAccessRoleArn=self.ROLE_ARN,
            InputDataConfig=self.INPUT_DATA_CONFIG,
            OutputDataConfig={"S3Uri": "s3://test/training_output/"},
            LanguageCode="en",
            Mode="MULTI_CLASS",
            VersionName="v1",
        )

    @pytest.mark.parametrize(
        "wait_for_completion, deferrable",
        [
            pytest.param(False, False, id="no_wait"),
            pytest.param(True, False, id="wait"),
            pytest.param(False, True, id="defer"),
        ],
    )
    @mock.patch.object(ComprehendHook, "get_waiter")
    def test_create_document_classifier_wait_combinations(
        self, _, wait_for_completion, deferrable, mock_conn, comprehend_hook
    ):
        self.operator.wait_for_completion = wait_for_completion
        self.operator.deferrable = deferrable

        response = self.operator.execute({})

        assert response == self.CLASSIFIER_ARN
        assert comprehend_hook.get_waiter.call_count == wait_for_completion
        assert self.operator.defer.call_count == deferrable

    def test_template_fields(self):
        validate_template_fields(self.operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.hooks.mwaa import MwaaHook
from airflow.providers.amazon.aws.hooks.sts import StsHook
from airflow.providers.amazon.aws.operators.mwaa import MwaaTriggerDagRunOperator
from airflow.providers.amazon.aws.sensors.mwaa import MwaaDagRunSensor

from system.amazon.aws.utils import SystemTestContextBuilder

DAG_ID = "example_mwaa"

# Externally fetched variables:
EXISTING_ENVIRONMENT_NAME_KEY = "ENVIRONMENT_NAME"
EXISTING_DAG_ID_KEY = "DAG_ID"
ROLE_WITHOUT_INVOKE_REST_API_ARN_KEY = "ROLE_WITHOUT_INVOKE_REST_API_ARN"

sys_test_context_task = (
    SystemTestContextBuilder()
    # NOTE: Creating a functional MWAA environment is time-consuming and requires
    # manually creating and configuring an S3 bucket for DAG storage and a VPC with
    # private subnets which is out of scope for this demo. To simplify this demo and
    # make it run in a reasonable time, an existing MWAA environment already
    # containing a DAG is required.
    # Here's a quick start guide to create an MWAA environment using AWS CloudFormation:
    # https://docs.aws.amazon.com/mwaa/latest/userguide/quick-start.html
    # If creating the environment using the AWS console, make sure to have a VPC with
    # at least 1 private subnet to be able to select the VPC while going through the
    # environment creation steps in the console wizard.
    # Make sure to set the environment variables with appropriate values
    .add_variable(EXISTING_ENVIRONMENT_NAME_KEY)
    .add_variable(EXISTING_DAG_ID_KEY)
    .add_variable(ROLE_WITHOUT_INVOKE_REST_API_ARN_KEY)
    .build()
)


@task
def unpause_dag(env_name: str, dag_id: str):
    mwaa_hook = MwaaHook()
    response = mwaa_hook.invoke_rest_api(
        env_name=env_name, path=f"/dags/{dag_id}", method="PATCH", body={"is_paused": False}
    )
    return not response["RestApiResponse"]["is_paused"]


# This task in the system test verifies that the MwaaHook's IAM fallback mechanism continues to work with
# the live MWAA API. This fallback depends on parsing a specific error message from the MWAA API, so we
# want to ensure we find out if the API response format ever changes. Unit tests cover this with mocked
# responses, but this system test validates against the real API.
@task
def test_iam_fallback(role_to_assume_arn, mwaa_env_name):
    assumed_role = StsHook().conn.assume_role(
        RoleArn=role_to_assume_arn, RoleSessionName="MwaaSysTestIamFallback"
    )

    credentials = assumed_role["Credentials"]
    session = boto3.Session(
        aws_access_key_id=credentials["AccessKeyId"],
        aws_secret_access_key=credentials["SecretAccessKey"],
        aws_session_token=credentials["SessionToken"],
    )

    mwaa_hook = MwaaHook()
    mwaa_hook.conn = session.client("mwaa")
    response = mwaa_hook.invoke_rest_api(env_name=mwaa_env_name, path="/dags", method="GET")
    return "dags" in response["RestApiResponse"]


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_name = test_context[EXISTING_ENVIRONMENT_NAME_KEY]
    trigger_dag_id = test_context[EXISTING_DAG_ID_KEY]
    restricted_role_arn = test_context[ROLE_WITHOUT_INVOKE_REST_API_ARN_KEY]

    # [START howto_operator_mwaa_trigger_dag_run]
    trigger_dag_run = MwaaTriggerDagRunOperator(
        task_id="trigger_dag_run",
        env_name=env_name,
        trigger_dag_id=trigger_dag_id,
        wait_for_completion=True,
    )
    # [END howto_operator_mwaa_trigger_dag_run]

    # [START howto_sensor_mwaa_dag_run]
    wait_for_dag_run = MwaaDagRunSensor(
        task_id="wait_for_dag_run",
        external_env_name=env_name,
        external_dag_id=trigger_dag_id,
        external_dag_run_id="{{ task_instance.xcom_pull(task_ids='trigger_dag_run')['RestApiResponse']['dag_run_id'] }}",
        poke_interval=5,
    )
    # [END howto_sensor_mwaa_dag_run]

    chain(
        # TEST SETUP
        test_context,
        # TEST BODY
        unpause_dag(env_name, trigger_dag_id),
        trigger_dag_run,
        wait_for_dag_run,
        test_iam_fallback(restricted_role_arn, env_name),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime as dt
import json
import random
from datetime import datetime

import boto3

from airflow import DAG, settings
from airflow.decorators import task, task_group
from airflow.models import Connection
from airflow.models.baseoperator import chain
from airflow.providers.amazon.aws.hooks.kinesis_analytics import KinesisAnalyticsV2Hook
from airflow.providers.amazon.aws.operators.kinesis_analytics import (
    KinesisAnalyticsV2CreateApplicationOperator,
    KinesisAnalyticsV2StartApplicationOperator,
    KinesisAnalyticsV2StopApplicationOperator,
)
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.kinesis_analytics import (
    KinesisAnalyticsV2StartApplicationCompletedSensor,
    KinesisAnalyticsV2StopApplicationCompletedSensor,
)
from airflow.providers.amazon.aws.transfers.http_to_s3 import HttpToS3Operator
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import SystemTestContextBuilder

ROLE_ARN_KEY = "ROLE_ARN"
sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

DAG_ID = "example_kinesis_analytics"


@task_group
def kinesis_analytics_v2_workflow():
    # [START howto_operator_create_application]
    create_application = KinesisAnalyticsV2CreateApplicationOperator(
        task_id="create_application",
        application_name=application_name,
        runtime_environment="FLINK-1_18",
        service_execution_role=test_context[ROLE_ARN_KEY],
        create_application_kwargs={
            "ApplicationConfiguration": {
                "FlinkApplicationConfiguration": {
                    "ParallelismConfiguration": {
                        "ConfigurationType": "CUSTOM",
                        "Parallelism": 2,
                        "ParallelismPerKPU": 1,
                        "AutoScalingEnabled": False,
                    }
                },
                "EnvironmentProperties": {
                    "PropertyGroups": [
                        {
                            "PropertyGroupId": "BlueprintMetadata",
                            "PropertyMap": {
                                "AWSRegion": region_name,
                                "BlueprintName": "KDS_FLINK-DATASTREAM-JAVA_S3",
                                "BucketName": f"s3://{bucket_name}/",
                                "PartitionFormat": "yyyy-MM-dd-HH",
                                "StreamInitialPosition": "TRIM_HORIZON",
                                "StreamName": stream_name,
                            },
                        },
                    ]
                },
                "ApplicationCodeConfiguration": {
                    "CodeContent": {
                        "S3ContentLocation": {
                            "BucketARN": f"arn:aws:s3:::{bucket_name}",
                            "FileKey": "code/kds-to-s3-datastream-java-1.0.1.jar",
                        },
                    },
                    "CodeContentType": "ZIPFILE",
                },
            }
        },
    )
    # [END howto_operator_create_application]

    # [START howto_operator_start_application]
    start_application = KinesisAnalyticsV2StartApplicationOperator(
        task_id="start_application",
        application_name=application_name,
    )
    # [END howto_operator_start_application]
    start_application.wait_for_completion = False

    # [START howto_sensor_start_application]
    await_start_application = KinesisAnalyticsV2StartApplicationCompletedSensor(
        task_id="await_start_application",
        application_name=application_name,
    )
    # [END howto_sensor_start_application]

    # [START howto_operator_stop_application]
    stop_application = KinesisAnalyticsV2StopApplicationOperator(
        task_id="stop_application",
        application_name=application_name,
    )
    # [END howto_operator_stop_application]
    stop_application.wait_for_completion = False
    # With the default `force=False` the test fails  because the data retention policy is blocking the deletion.
    stop_application.force = True

    # [START howto_sensor_stop_application]
    await_stop_application = KinesisAnalyticsV2StopApplicationCompletedSensor(
        task_id="await_stop_application",
        application_name=application_name,
    )
    # [END howto_sensor_stop_application]

    @task(trigger_rule=TriggerRule.ALL_DONE)
    def delete_application(app_name: str):
        kinesis_analytics_v2_hook = KinesisAnalyticsV2Hook()
        response = kinesis_analytics_v2_hook.conn.describe_application(ApplicationName=app_name)
        kinesis_analytics_v2_hook.conn.delete_application(
            ApplicationName=app_name, CreateTimestamp=response["ApplicationDetail"]["CreateTimestamp"]
        )

    chain(
        create_application,
        start_application,
        await_start_application,
        stop_application,
        await_stop_application,
        delete_application(application_name),
    )


@task_group
def copy_jar_to_s3(bucket: str):
    """

    Copy application code to S3 using HttpToS3Operator.

    :param bucket: Name of the Amazon S3 bucket.
    """

    @task
    def create_connection(conn_id):
        conn = Connection(
            conn_id=conn_id,
            conn_type="http",
            host="https://github.com/",
        )
        session = settings.Session()
        session.add(conn)
        session.commit()

    @task(trigger_rule=TriggerRule.ALL_DONE)
    def delete_connection(conn_id: str):
        session = settings.Session()
        conn_to_details = session.query(Connection).filter(Connection.conn_id == conn_id).first()
        session.delete(conn_to_details)
        session.commit()

    copy_jar_file = HttpToS3Operator(
        task_id="copy_jar_file",
        http_conn_id=http_conn_id,
        endpoint="awslabs/managed-service-for-apache-flink-blueprints/releases/download/v2.0.1/kds-to-s3-datastream-java-1.0.1.jar",
        s3_bucket=bucket,
        s3_key="code/kds-to-s3-datastream-java-1.0.1.jar",
    )

    chain(create_connection(http_conn_id), copy_jar_file, delete_connection(http_conn_id))


@task
def create_kinesis_stream(stream: str, region: str):
    """
    Create kinesis stream and put some sample data.

    :param stream: Name of the kinesis stream.
    :param region: Region name
    """
    client = boto3.client("kinesis", region_name=region)
    client.create_stream(StreamName=stream, ShardCount=1, StreamModeDetails={"StreamMode": "PROVISIONED"})
    account_id = boto3.client("sts").get_caller_identity()["Account"]
    waiter = client.get_waiter("stream_exists")
    waiter.wait(StreamName=stream, WaiterConfig={"Delay": 60, "MaxAttempts": 4})

    def get_data():
        return {
            "event_time": dt.datetime.now().isoformat(),
            "ticker": random.choice(["AAPL", "AMZN", "MSFT", "INTC", "TBV"]),
            "price": round(random.random() * 100, 2),
        }

    for _ in range(2):
        data = get_data()
        client.put_record(
            StreamARN=f"arn:aws:kinesis:{region}:{account_id}:stream/{stream}",
            Data=json.dumps(data),
            PartitionKey=data["ticker"],
        )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_kinesis_stream(stream: str, region: str):
    client = boto3.client("kinesis", region_name=region)
    client.delete_stream(StreamName=stream, EnforceConsumerDeletion=True)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]
    bucket_name = f"{env_id}-kinesis-analytics"
    application_name = f"{env_id}-test-app"
    http_conn_id = f"{env_id}-git"
    region_name = boto3.session.Session().region_name
    stream_name = f"{env_id}-test-stream"

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=bucket_name,
    )

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=bucket_name,
        force_delete=True,
    )

    chain(
        # TEST SETUP
        test_context,
        create_bucket,
        create_kinesis_stream(stream=stream_name, region=region_name),
        copy_jar_to_s3(bucket=bucket_name),
        # TEST BODY
        kinesis_analytics_v2_workflow(),
        # TEST TEARDOWN
        delete_kinesis_stream(stream=stream_name, region=region_name),
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import pytest

from airflow.providers.amazon.aws.utils.tags import format_tags


@pytest.fixture
def source():
    return {"key1": "value1", "key2": "value2"}


def test_format_tags_with_dict(source):
    formatted_tags = format_tags(source)
    expected_result = [{"Key": "key1", "Value": "value1"}, {"Key": "key2", "Value": "value2"}]
    assert formatted_tags == expected_result


def test_format_tags_with_other_input():
    source = [{"Key": "key1", "Value": "value1"}, {"Key": "key2", "Value": "value2"}]
    formatted_tags = format_tags(source)
    assert formatted_tags == source


def test_format_tags_with_custom_labels(source):
    formatted_tags = format_tags(source, key_label="TagKey", value_label="TagValue")
    expected_result = [{"TagKey": "key1", "TagValue": "value1"}, {"TagKey": "key2", "TagValue": "value2"}]
    assert formatted_tags == expected_result


def test_format_tags_with_none_input():
    formatted_tags = format_tags(None)
    assert formatted_tags == []



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import boto3
import pytest
from botocore.exceptions import WaiterError

from airflow.providers.amazon.aws.hooks.dynamodb import DynamoDBHook

TEST_IMPORT_ARN = "arn:aws:dynamodb:us-east-1:255683865591:table/test-table/import/01662190284205-aa94decf"


class TestCustomDynamoDBServiceWaiters:
    EXPORT_STATUS_COMPLETED = "COMPLETED"
    EXPORT_STATUS_FAILED = "FAILED"
    EXPORT_STATUS_IN_PROGRESS = "IN_PROGRESS"

    IMPORT_STATUS_FAILED = ("CANCELLING", "CANCELLED", "FAILED")
    IMPORT_STATUS_COMPLETED = "COMPLETED"
    IMPORT_STATUS_IN_PROGRESS = "IN_PROGRESS"

    @pytest.fixture(autouse=True)
    def _setup_test_cases(self, monkeypatch):
        self.resource = boto3.resource("dynamodb", region_name="eu-west-3")
        monkeypatch.setattr(DynamoDBHook, "conn", self.resource)
        self.client = self.resource.meta.client

    @pytest.fixture
    def mock_describe_export(self):
        """Mock ``DynamoDBHook.Client.describe_export`` method."""
        with mock.patch.object(self.client, "describe_export") as m:
            yield m

    @pytest.fixture
    def mock_describe_import(self):
        """Mock ``DynamoDBHook.Client.describe_import`` method."""
        with mock.patch.object(self.client, "describe_import") as m:
            yield m

    def test_service_waiters(self):
        hook_waiters = DynamoDBHook(aws_conn_id=None).list_waiters()
        assert "export_table" in hook_waiters
        assert "import_table" in hook_waiters

    @staticmethod
    def describe_export(status: str):
        """
        Helper function for generate minimal DescribeExport response for single job.
        https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_DescribeExport.html
        """
        return {"ExportDescription": {"ExportStatus": status}}

    @staticmethod
    def describe_import(status: str):
        """
        Helper function for generate minimal DescribeImport response for single job.
        https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_DescribeImport.html
        """
        return {"ImportTableDescription": {"ImportStatus": status}}

    def test_export_table_to_point_in_time_completed(self, mock_describe_export):
        """Test state transition from `in progress` to `completed` during init."""
        waiter = DynamoDBHook(aws_conn_id=None).get_waiter("export_table")
        mock_describe_export.side_effect = [
            self.describe_export(self.EXPORT_STATUS_IN_PROGRESS),
            self.describe_export(self.EXPORT_STATUS_COMPLETED),
        ]
        waiter.wait(
            ExportArn="LoremIpsumissimplydummytextoftheprintingandtypesettingindustry",
            WaiterConfig={"Delay": 0.01, "MaxAttempts": 3},
        )

    def test_export_table_to_point_in_time_failed(self, mock_describe_export):
        """Test state transition from `in progress` to `failed` during init."""
        with mock.patch("boto3.client") as client:
            client.return_value = self.client
            mock_describe_export.side_effect = [
                self.describe_export(self.EXPORT_STATUS_IN_PROGRESS),
                self.describe_export(self.EXPORT_STATUS_FAILED),
            ]
            waiter = DynamoDBHook(aws_conn_id=None).get_waiter("export_table", client=self.client)
            with pytest.raises(WaiterError, match='we matched expected path: "FAILED"'):
                waiter.wait(
                    ExportArn="LoremIpsumissimplydummytextoftheprintingandtypesettingindustry",
                    WaiterConfig={"Delay": 0.01, "MaxAttempts": 3},
                )

    def test_import_table_completed(self, mock_describe_import):
        waiter = DynamoDBHook(aws_conn_id=None).get_waiter("import_table")
        mock_describe_import.side_effect = [
            self.describe_import(self.IMPORT_STATUS_IN_PROGRESS),
            self.describe_import(self.IMPORT_STATUS_COMPLETED),
        ]
        waiter.wait(
            ImportArn=TEST_IMPORT_ARN,
            WaiterConfig={"Delay": 0.01, "MaxAttempts": 3},
        )

    @pytest.mark.parametrize(
        "status",
        [
            pytest.param(IMPORT_STATUS_FAILED[0]),
            pytest.param(IMPORT_STATUS_FAILED[1]),
            pytest.param(IMPORT_STATUS_FAILED[2]),
        ],
    )
    def test_import_table_failed(self, status, mock_describe_import):
        waiter = DynamoDBHook(aws_conn_id=None).get_waiter("import_table")
        mock_describe_import.side_effect = [
            self.describe_import(self.EXPORT_STATUS_IN_PROGRESS),
            self.describe_import(status),
        ]
        with pytest.raises(WaiterError, match=f'we matched expected path: "{status}"'):
            waiter.wait(ImportArn=TEST_IMPORT_ARN, WaiterConfig={"Delay": 0.01, "MaxAttempts": 3})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.hooks.opensearch_serverless import OpenSearchServerlessHook
from airflow.providers.amazon.aws.triggers.base import AwsBaseWaiterTrigger
from airflow.utils.helpers import exactly_one

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook


class OpenSearchServerlessCollectionActiveTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when an Amazon OpenSearch Serverless Collection reaches the ACTIVE state.

    :param collection_id: A collection ID. You can't provide a name and an ID in the same request.
    :param collection_name: A collection name. You can't provide a name and an ID in the same request.

    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 60)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 20)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        *,
        collection_id: str | None = None,
        collection_name: str | None = None,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 20,
        aws_conn_id: str | None = None,
    ) -> None:
        if not exactly_one(collection_id is None, collection_name is None):
            raise AttributeError("Either collection_ids or collection_names must be provided, not both.")

        super().__init__(
            serialized_fields={"collection_id": collection_id, "collection_name": collection_name},
            waiter_name="collection_available",
            # waiter_args is a dict[str, Any], allow a possible list of None (it is caught above)
            waiter_args={"ids": [collection_id]} if collection_id else {"names": [collection_name]},  # type: ignore[list-item]
            failure_message="OpenSearch Serverless Collection creation failed.",
            status_message="Status of OpenSearch Serverless Collection is",
            status_queries=["status"],
            return_key="collection_id" if collection_id else "collection_name",
            return_value=collection_id if collection_id else collection_name,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return OpenSearchServerlessHook(aws_conn_id=self.aws_conn_id)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from botocore.exceptions import ClientError

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.eks import EksHook
from airflow.providers.amazon.aws.triggers.base import AwsBaseWaiterTrigger
from airflow.providers.amazon.aws.utils.waiter_with_logging import async_wait
from airflow.triggers.base import TriggerEvent

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook


class EksCreateClusterTrigger(AwsBaseWaiterTrigger):
    """
    Trigger for EksCreateClusterOperator.

    The trigger will asynchronously wait for the cluster to be created.

    :param cluster_name: The name of the EKS cluster
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    :param region_name: Which AWS region the connection should use.
         If this is None or empty then the default boto3 behaviour is used.
    """

    def __init__(
        self,
        cluster_name: str,
        waiter_delay: int,
        waiter_max_attempts: int,
        aws_conn_id: str | None,
        region_name: str | None = None,
    ):
        super().__init__(
            serialized_fields={"cluster_name": cluster_name, "region_name": region_name},
            waiter_name="cluster_active",
            waiter_args={"name": cluster_name},
            failure_message="Error checking Eks cluster",
            status_message="Eks cluster status is",
            status_queries=["cluster.status"],
            return_value=None,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
            region_name=region_name,
        )

    def hook(self) -> AwsGenericHook:
        return EksHook(aws_conn_id=self.aws_conn_id, region_name=self.region_name)

    async def run(self):
        async with await self.hook().get_async_conn() as client:
            waiter = client.get_waiter(self.waiter_name)
            try:
                await async_wait(
                    waiter,
                    self.waiter_delay,
                    self.attempts,
                    self.waiter_args,
                    self.failure_message,
                    self.status_message,
                    self.status_queries,
                )
            except AirflowException as exception:
                self.log.error("Error creating cluster: %s", exception)
                yield TriggerEvent({"status": "failed"})
            else:
                yield TriggerEvent({"status": "success"})


class EksDeleteClusterTrigger(AwsBaseWaiterTrigger):
    """
    Trigger for EksDeleteClusterOperator.

    The trigger will asynchronously wait for the cluster to be deleted. If there are
    any nodegroups or fargate profiles associated with the cluster, they will be deleted
    before the cluster is deleted.

    :param cluster_name: The name of the EKS cluster
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    :param region_name: Which AWS region the connection should use.
         If this is None or empty then the default boto3 behaviour is used.
    :param force_delete_compute: If True, any nodegroups or fargate profiles associated
        with the cluster will be deleted before the cluster is deleted.
    """

    def __init__(
        self,
        cluster_name,
        waiter_delay: int,
        waiter_max_attempts: int,
        aws_conn_id: str | None,
        region_name: str | None,
        force_delete_compute: bool,
    ):
        self.cluster_name = cluster_name
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts
        self.aws_conn_id = aws_conn_id
        self.region_name = region_name
        self.force_delete_compute = force_delete_compute

    def serialize(self) -> tuple[str, dict[str, Any]]:
        return (
            self.__class__.__module__ + "." + self.__class__.__qualname__,
            {
                "cluster_name": self.cluster_name,
                "waiter_delay": str(self.waiter_delay),
                "waiter_max_attempts": str(self.waiter_max_attempts),
                "aws_conn_id": self.aws_conn_id,
                "region_name": self.region_name,
                "force_delete_compute": self.force_delete_compute,
            },
        )

    def hook(self) -> AwsGenericHook:
        return EksHook(aws_conn_id=self.aws_conn_id, region_name=self.region_name)

    async def run(self):
        async with await self.hook().get_async_conn() as client:
            waiter = client.get_waiter("cluster_deleted")
            if self.force_delete_compute:
                await self.delete_any_nodegroups(client=client)
                await self.delete_any_fargate_profiles(client=client)
            try:
                await client.delete_cluster(name=self.cluster_name)
            except ClientError as ex:
                if ex.response.get("Error").get("Code") == "ResourceNotFoundException":
                    pass
                else:
                    raise
            await async_wait(
                waiter=waiter,
                waiter_delay=int(self.waiter_delay),
                waiter_max_attempts=int(self.waiter_max_attempts),
                args={"name": self.cluster_name},
                failure_message="Error deleting cluster",
                status_message="Status of cluster is",
                status_args=["cluster.status"],
            )

        yield TriggerEvent({"status": "deleted"})

    async def delete_any_nodegroups(self, client) -> None:
        """
        Delete all EKS Nodegroups for a provided Amazon EKS Cluster.

        All the EKS Nodegroups are deleted simultaneously. We wait for
        all Nodegroups to be deleted before returning.
        """
        nodegroups = await client.list_nodegroups(clusterName=self.cluster_name)
        if nodegroups.get("nodegroups", None):
            self.log.info("Deleting nodegroups")
            waiter = self.hook().get_waiter("all_nodegroups_deleted", deferrable=True, client=client)
            for group in nodegroups["nodegroups"]:
                await client.delete_nodegroup(clusterName=self.cluster_name, nodegroupName=group)
            await async_wait(
                waiter=waiter,
                waiter_delay=int(self.waiter_delay),
                waiter_max_attempts=int(self.waiter_max_attempts),
                args={"clusterName": self.cluster_name},
                failure_message=f"Error deleting nodegroup for cluster {self.cluster_name}",
                status_message="Deleting nodegroups associated with the cluster",
                status_args=["nodegroups"],
            )
            self.log.info("All nodegroups deleted")
        else:
            self.log.info("No nodegroups associated with cluster %s", self.cluster_name)

    async def delete_any_fargate_profiles(self, client) -> None:
        """
        Delete all EKS Fargate profiles for a provided Amazon EKS Cluster.

        EKS Fargate profiles must be deleted one at a time, so we must wait
        for one to be deleted before sending the next delete command.
        """
        fargate_profiles = await client.list_fargate_profiles(clusterName=self.cluster_name)
        if fargate_profiles.get("fargateProfileNames"):
            self.log.info("Waiting for Fargate profiles to delete.  This will take some time.")
            for profile in fargate_profiles["fargateProfileNames"]:
                await client.delete_fargate_profile(clusterName=self.cluster_name, fargateProfileName=profile)
                await async_wait(
                    waiter=client.get_waiter("fargate_profile_deleted"),
                    waiter_delay=int(self.waiter_delay),
                    waiter_max_attempts=int(self.waiter_max_attempts),
                    args={"clusterName": self.cluster_name, "fargateProfileName": profile},
                    failure_message=f"Error deleting fargate profile for cluster {self.cluster_name}",
                    status_message="Status of fargate profile is",
                    status_args=["fargateProfile.status"],
                )
            self.log.info("All Fargate profiles deleted")
        else:
            self.log.info("No Fargate profiles associated with cluster %s", self.cluster_name)


class EksCreateFargateProfileTrigger(AwsBaseWaiterTrigger):
    """
    Asynchronously wait for the fargate profile to be created.

    :param cluster_name: The name of the EKS cluster
    :param fargate_profile_name: The name of the fargate profile
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        cluster_name: str,
        fargate_profile_name: str,
        waiter_delay: int,
        waiter_max_attempts: int,
        aws_conn_id: str | None,
        region_name: str | None = None,
    ):
        super().__init__(
            serialized_fields={"cluster_name": cluster_name, "fargate_profile_name": fargate_profile_name},
            waiter_name="fargate_profile_active",
            waiter_args={"clusterName": cluster_name, "fargateProfileName": fargate_profile_name},
            failure_message="Failure while creating Fargate profile",
            status_message="Fargate profile not created yet",
            status_queries=["fargateProfile.status"],
            return_value=None,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
            region_name=region_name,
        )

    def hook(self) -> AwsGenericHook:
        return EksHook(aws_conn_id=self.aws_conn_id, region_name=self.region_name)


class EksDeleteFargateProfileTrigger(AwsBaseWaiterTrigger):
    """
    Asynchronously wait for the fargate profile to be deleted.

    :param cluster_name: The name of the EKS cluster
    :param fargate_profile_name: The name of the fargate profile
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        cluster_name: str,
        fargate_profile_name: str,
        waiter_delay: int,
        waiter_max_attempts: int,
        aws_conn_id: str | None,
        region_name: str | None = None,
    ):
        super().__init__(
            serialized_fields={"cluster_name": cluster_name, "fargate_profile_name": fargate_profile_name},
            waiter_name="fargate_profile_deleted",
            waiter_args={"clusterName": cluster_name, "fargateProfileName": fargate_profile_name},
            failure_message="Failure while deleting Fargate profile",
            status_message="Fargate profile not deleted yet",
            status_queries=["fargateProfile.status"],
            return_value=None,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
            region_name=region_name,
        )

    def hook(self) -> AwsGenericHook:
        return EksHook(aws_conn_id=self.aws_conn_id, region_name=self.region_name)


class EksCreateNodegroupTrigger(AwsBaseWaiterTrigger):
    """
    Trigger for EksCreateNodegroupOperator.

    The trigger will asynchronously poll the boto3 API and wait for the
    nodegroup to be in the state specified by the waiter.

    :param cluster_name: The name of the EKS cluster associated with the node group.
    :param nodegroup_name: The name of the nodegroup to check.
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    :param region_name: Which AWS region the connection should use. (templated)
        If this is None or empty then the default boto3 behaviour is used.
    """

    def __init__(
        self,
        cluster_name: str,
        nodegroup_name: str,
        waiter_delay: int,
        waiter_max_attempts: int,
        aws_conn_id: str | None,
        region_name: str | None = None,
    ):
        super().__init__(
            serialized_fields={
                "cluster_name": cluster_name,
                "nodegroup_name": nodegroup_name,
                "region_name": region_name,
            },
            waiter_name="nodegroup_active",
            waiter_args={"clusterName": cluster_name, "nodegroupName": nodegroup_name},
            failure_message="Error creating nodegroup",
            status_message="Nodegroup status is",
            status_queries=["nodegroup.status", "nodegroup.health.issues"],
            return_value=None,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
            region_name=region_name,
        )

    def hook(self) -> AwsGenericHook:
        return EksHook(aws_conn_id=self.aws_conn_id, region_name=self.region_name)


class EksDeleteNodegroupTrigger(AwsBaseWaiterTrigger):
    """
    Trigger for EksDeleteNodegroupOperator.

    The trigger will asynchronously poll the boto3 API and wait for the
    nodegroup to be in the state specified by the waiter.

    :param cluster_name: The name of the EKS cluster associated with the node group.
    :param nodegroup_name: The name of the nodegroup to check.
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    :param region_name: Which AWS region the connection should use. (templated)
        If this is None or empty then the default boto3 behaviour is used.
    """

    def __init__(
        self,
        cluster_name: str,
        nodegroup_name: str,
        waiter_delay: int,
        waiter_max_attempts: int,
        aws_conn_id: str | None,
        region_name: str | None = None,
    ):
        super().__init__(
            serialized_fields={"cluster_name": cluster_name, "nodegroup_name": nodegroup_name},
            waiter_name="nodegroup_deleted",
            waiter_args={"clusterName": cluster_name, "nodegroupName": nodegroup_name},
            failure_message="Error deleting nodegroup",
            status_message="Nodegroup status is",
            status_queries=["nodegroup.status", "nodegroup.health.issues"],
            return_value=None,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
            region_name=region_name,
        )

    def hook(self) -> AwsGenericHook:
        return EksHook(aws_conn_id=self.aws_conn_id, region_name=self.region_name)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.glue import GlueDataQualityHook, GlueJobHook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.triggers.glue import (
    GlueDataQualityRuleRecommendationRunCompleteTrigger,
    GlueDataQualityRuleSetEvaluationRunCompleteTrigger,
)
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.sensors.base import BaseSensorOperator

if TYPE_CHECKING:
    from airflow.utils.context import Context


class GlueJobSensor(BaseSensorOperator):
    """
    Waits for an AWS Glue Job to reach any of the status below.

    'FAILED', 'STOPPED', 'SUCCEEDED'

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:GlueJobSensor`

    :param job_name: The AWS Glue Job unique name
    :param run_id: The AWS Glue current running job identifier
    :param verbose: If True, more Glue Job Run logs show in the Airflow Task Logs.  (default: False)
    """

    template_fields: Sequence[str] = ("job_name", "run_id")

    def __init__(
        self,
        *,
        job_name: str,
        run_id: str,
        verbose: bool = False,
        aws_conn_id: str | None = "aws_default",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.job_name = job_name
        self.run_id = run_id
        self.verbose = verbose
        self.aws_conn_id = aws_conn_id
        self.success_states: list[str] = ["SUCCEEDED"]
        self.errored_states: list[str] = ["FAILED", "STOPPED", "TIMEOUT"]
        self.next_log_tokens = GlueJobHook.LogContinuationTokens()

    @cached_property
    def hook(self):
        return GlueJobHook(aws_conn_id=self.aws_conn_id)

    def poke(self, context: Context):
        self.log.info("Poking for job run status :for Glue Job %s and ID %s", self.job_name, self.run_id)
        job_state = self.hook.get_job_state(job_name=self.job_name, run_id=self.run_id)

        try:
            if job_state in self.success_states:
                self.log.info("Exiting Job %s Run State: %s", self.run_id, job_state)
                return True
            if job_state in self.errored_states:
                job_error_message = "Exiting Job %s Run State: %s", self.run_id, job_state
                self.log.info(job_error_message)
                raise AirflowException(job_error_message)
            return False
        finally:
            if self.verbose:
                self.hook.print_job_logs(
                    job_name=self.job_name,
                    run_id=self.run_id,
                    continuation_tokens=self.next_log_tokens,
                )


class GlueDataQualityRuleSetEvaluationRunSensor(AwsBaseSensor[GlueDataQualityHook]):
    """
    Waits for an AWS Glue data quality ruleset evaluation run to reach any of the status below.

    'FAILED', 'STOPPED', 'STOPPING', 'TIMEOUT', 'SUCCEEDED'

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:GlueDataQualityRuleSetEvaluationRunSensor`

    :param evaluation_run_id: The AWS Glue data quality ruleset evaluation run identifier.
    :param verify_result_status: Validate all the ruleset rules evaluation run results,
        If any of the rule status is Fail or Error then an exception is thrown. (default: True)
    :param show_results: Displays all the ruleset rules evaluation run results. (default: True)
    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 120)
    :param max_retries: Number of times before returning the current state. (default: 60)

    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    SUCCESS_STATES = ("SUCCEEDED",)

    FAILURE_STATES = ("FAILED", "STOPPED", "STOPPING", "TIMEOUT")

    aws_hook_class = GlueDataQualityHook
    template_fields: Sequence[str] = aws_template_fields("evaluation_run_id")

    def __init__(
        self,
        *,
        evaluation_run_id: str,
        show_results: bool = True,
        verify_result_status: bool = True,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poke_interval: int = 120,
        max_retries: int = 60,
        aws_conn_id: str | None = "aws_default",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.evaluation_run_id = evaluation_run_id
        self.show_results = show_results
        self.verify_result_status = verify_result_status
        self.aws_conn_id = aws_conn_id
        self.max_retries = max_retries
        self.poke_interval = poke_interval
        self.deferrable = deferrable

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=GlueDataQualityRuleSetEvaluationRunCompleteTrigger(
                    evaluation_run_id=self.evaluation_run_id,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="execute_complete",
            )
        else:
            super().execute(context=context)

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> None:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            message = f"Error: AWS Glue data quality ruleset evaluation run: {validated_event}"
            raise AirflowException(message)

        self.hook.validate_evaluation_run_results(
            evaluation_run_id=validated_event["evaluation_run_id"],
            show_results=self.show_results,
            verify_result_status=self.verify_result_status,
        )

        self.log.info("AWS Glue data quality ruleset evaluation run completed.")

    def poke(self, context: Context):
        self.log.info(
            "Poking for AWS Glue data quality ruleset evaluation run RunId: %s", self.evaluation_run_id
        )

        response = self.hook.conn.get_data_quality_ruleset_evaluation_run(RunId=self.evaluation_run_id)

        status = response.get("Status")

        if status in self.SUCCESS_STATES:
            self.hook.validate_evaluation_run_results(
                evaluation_run_id=self.evaluation_run_id,
                show_results=self.show_results,
                verify_result_status=self.verify_result_status,
            )

            self.log.info(
                "AWS Glue data quality ruleset evaluation run completed RunId: %s Run State: %s",
                self.evaluation_run_id,
                response["Status"],
            )

            return True

        if status in self.FAILURE_STATES:
            job_error_message = (
                f"Error: AWS Glue data quality ruleset evaluation run RunId: {self.evaluation_run_id} Run "
                f"Status: {status}"
                f": {response.get('ErrorString')}"
            )
            self.log.info(job_error_message)
            raise AirflowException(job_error_message)
        return False


class GlueDataQualityRuleRecommendationRunSensor(AwsBaseSensor[GlueDataQualityHook]):
    """
    Waits for an AWS Glue data quality recommendation run to reach any of the status below.

    'FAILED', 'STOPPED', 'STOPPING', 'TIMEOUT', 'SUCCEEDED'

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:GlueDataQualityRuleRecommendationRunSensor`

    :param recommendation_run_id: The AWS Glue data quality rule recommendation run identifier.
    :param show_results: Displays the recommended ruleset (a set of rules), when recommendation run completes. (default: True)
    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 120)
    :param max_retries: Number of times before returning the current state. (default: 60)

    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    SUCCESS_STATES = ("SUCCEEDED",)

    FAILURE_STATES = ("FAILED", "STOPPED", "STOPPING", "TIMEOUT")

    aws_hook_class = GlueDataQualityHook
    template_fields: Sequence[str] = aws_template_fields("recommendation_run_id")

    def __init__(
        self,
        *,
        recommendation_run_id: str,
        show_results: bool = True,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poke_interval: int = 120,
        max_retries: int = 60,
        aws_conn_id: str | None = "aws_default",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.recommendation_run_id = recommendation_run_id
        self.show_results = show_results
        self.deferrable = deferrable
        self.poke_interval = poke_interval
        self.max_retries = max_retries
        self.aws_conn_id = aws_conn_id

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=GlueDataQualityRuleRecommendationRunCompleteTrigger(
                    recommendation_run_id=self.recommendation_run_id,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="execute_complete",
            )
        else:
            super().execute(context=context)

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> None:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            message = f"Error: AWS Glue data quality recommendation run: {validated_event}"
            raise AirflowException(message)

        if self.show_results:
            self.hook.log_recommendation_results(run_id=self.recommendation_run_id)

        self.log.info("AWS Glue data quality recommendation run completed.")

    def poke(self, context: Context) -> bool:
        self.log.info(
            "Poking for AWS Glue data quality recommendation run RunId: %s", self.recommendation_run_id
        )

        response = self.hook.conn.get_data_quality_rule_recommendation_run(RunId=self.recommendation_run_id)

        status = response.get("Status")

        if status in self.SUCCESS_STATES:
            if self.show_results:
                self.hook.log_recommendation_results(run_id=self.recommendation_run_id)

            self.log.info(
                "AWS Glue data quality recommendation run completed RunId: %s Run State: %s",
                self.recommendation_run_id,
                response["Status"],
            )

            return True

        if status in self.FAILURE_STATES:
            job_error_message = (
                f"Error: AWS Glue data quality recommendation run RunId: {self.recommendation_run_id} Run "
                f"Status: {status}"
                f": {response.get('ErrorString')}"
            )
            self.log.info(job_error_message)
            raise AirflowException(job_error_message)
        return False
